import numpy as np
from meteva.base.tool.math_tools import sxy_iteration,ss_iteration
from meteva.base import IV
def sample_count(Ob, Fo=None):
    '''
    计算检验的样本数
    -----------------------------
    :param Ob: 实况数据  任意维numpy数组
    :param Fo: 预测数据 任意维numpy数组,Fo.shape 和Ob.shape一致
    :return: 整数，Ob.size
    '''
    return Ob.size


def ob_fo_sum(ob,fo):
    Fo_shape = fo.shape
    Ob_shape = ob.shape

    Ob_shpe_list = list(Ob_shape)
    size = len(Ob_shpe_list)
    ind = -size
    Fo_Ob_index = list(Fo_shape[ind:])
    if Fo_Ob_index != Ob_shpe_list:
        print('预报数据和观测数据维度不匹配')
        return

    if len(fo.shape) == len(ob.shape):
        result = [np.sum(ob),np.sum(fo)]
    else:
        result = [np.sum(ob)]
        for i in range(Fo_shape[0]):
            result.append(np.sum(fo[i,:]))
    result = np.array(result)
    return result

def ob_fo_mean(ob,fo):

    Fo_shape = fo.shape
    Ob_shape = ob.shape

    Ob_shpe_list = list(Ob_shape)
    size = len(Ob_shpe_list)
    ind = -size
    Fo_Ob_index = list(Fo_shape[ind:])
    if Fo_Ob_index != Ob_shpe_list:
        print('预报数据和观测数据维度不匹配')
        return

    if len(fo.shape) == len(ob.shape):
        result = [np.mean(ob),np.mean(fo)]
    else:
        result = [np.mean(ob)]
        for i in range(Fo_shape[0]):
            result.append(np.mean(fo[i,:]))
    result = np.array(result)
    return result


def ob_quantile(ob,fo,grade_list=[0.5]):
    '''
    统计观测数据的百分位
    :param ob:
    :param fo:
    :param grade_list:
    :return:
    '''
    ob1 = np.sort(ob)
    ob1 = ob1.flatten()
    ob1.sort()
    n = len(ob1)
    qu_list = []
    for i in range(len(grade_list)):
        m = int(n * grade_list[i])
        qu_list.append(ob1[m])
    if len(qu_list) ==1:
        return qu_list[0]
    else:
        return np.array(qu_list)

def ob_fo_quantile(ob,fo,grade_list=[0.5]):
    '''
    统计观测数据的百分位
    :param ob:
    :param fo:
    :param grade_list:
    :return:
    '''


    Fo_shape = fo.shape
    Ob_shape = ob.shape

    Ob_shpe_list = list(Ob_shape)
    size = len(Ob_shpe_list)
    ind = -size
    Fo_Ob_index = list(Fo_shape[ind:])

    if Fo_Ob_index != Ob_shpe_list:
        print('预报数据和观测数据维度不匹配')
        return
    if len(fo.shape) == len(ob.shape):
        ob1 = np.sort(ob)
        ob1 = ob1.flatten()
        ob1.sort()
        n = len(ob1)
        qu_list_ob = []
        for i in range(len(grade_list)):
            m = int(n * grade_list[i])
            qu_list_ob.append([ob1[m]])

        fo1 = np.sort(fo)
        fo1 = fo1.flatten()
        fo1.sort()
        qu_list_fo = []
        for i in range(len(grade_list)):
            m = int(n * grade_list[i])
            qu_list_fo.append([fo1[m]])

        qu_list = [qu_list_ob,qu_list_fo]
    else:
        ob1 = np.sort(ob)
        ob1 = ob1.flatten()
        ob1.sort()
        n = len(ob1)
        qu_list_ob = []
        for i in range(len(grade_list)):
            m = int(n * grade_list[i])
            qu_list_ob.append([ob1[m]])

        qu_list = [qu_list_ob]
        for k in range(Fo_shape[0]):
            fo1 = np.sort(fo[k,:])
            fo1 = fo1.flatten()
            fo1.sort()
            qu_list_fo = []
            for i in range(len(grade_list)):
                m = int(n * grade_list[i])
                qu_list_fo.append([fo1[m]])
            qu_list.append(qu_list_fo)

    result = np.array(qu_list).squeeze()
    return result



def ob_fo_min(ob,fo,count = 1):
    if count ==1:
        Fo_shape = fo.shape
        Ob_shape = ob.shape

        Ob_shpe_list = list(Ob_shape)
        size = len(Ob_shpe_list)
        ind = -size
        Fo_Ob_index = list(Fo_shape[ind:])
        if Fo_Ob_index != Ob_shpe_list:
            print('预报数据和观测数据维度不匹配')
            return

        if len(fo.shape) == len(ob.shape):
            result = [np.min(ob),np.min(fo)]
        else:
            result = [np.min(ob)]
            for i in range(Fo_shape[0]):
                result.append(np.min(fo[i,:]))
        result = np.array(result)
        return result
    elif count < 1:
        print('para count must be int >=1')
    else:
        Fo_shape = fo.shape
        Ob_shape = ob.shape

        Ob_shpe_list = list(Ob_shape)
        size = len(Ob_shpe_list)
        ind = -size
        Fo_Ob_index = list(Fo_shape[ind:])
        if Fo_Ob_index != Ob_shpe_list:
            print('预报数据和观测数据维度不匹配')
            return
        ob_f = ob.flatten()
        index = ob_f.argsort()[:count]
        ob_mins = ob_f[index]
        result = [ob_mins]
        if len(fo.shape) == len(ob.shape):
            fo_f = fo.flatten()
            index = fo_f.argsort()[:count]
            fo_maxs = fo_f[index]
            result.append(fo_maxs)
        else:
            for i in range(Fo_shape[0]):
                v1 = fo[i, :].flatten()
                index = v1.argsort()[:count]
                fo_maxs = v1[index]
                result.append(fo_maxs)
        result = np.array(result)
        return result

def ob_fo_max(ob,fo, count = 1):
    if count == 1:
        Fo_shape = fo.shape
        Ob_shape = ob.shape

        Ob_shpe_list = list(Ob_shape)
        size = len(Ob_shpe_list)
        ind = -size
        Fo_Ob_index = list(Fo_shape[ind:])
        if Fo_Ob_index != Ob_shpe_list:
            print('预报数据和观测数据维度不匹配')
            return

        if len(fo.shape) == len(ob.shape):
            result = [np.max(ob),np.max(fo)]
        else:
            result = [np.max(ob)]
            for i in range(Fo_shape[0]):
                result.append(np.max(fo[i,:]))
        result = np.array(result)
        return result
    elif count < 1:
        print('para count must be int >=1')
    else:
        Fo_shape = fo.shape
        Ob_shape = ob.shape

        Ob_shpe_list = list(Ob_shape)
        size = len(Ob_shpe_list)
        ind = -size
        Fo_Ob_index = list(Fo_shape[ind:])
        if Fo_Ob_index != Ob_shpe_list:
            print('预报数据和观测数据维度不匹配')
            return
        ob_f = ob.flatten()
        index = ob_f.argsort()[-count:][::-1]
        ob_maxs = ob_f[index]
        result = [ob_maxs]
        if len(fo.shape) == len(ob.shape):
            fo_f = fo.flatten()
            index = fo_f.argsort()[-count:][::-1]
            fo_maxs = fo_f[index]
            result.append(fo_maxs)
        else:
            for i in range(Fo_shape[0]):
                v1 = fo[i, :].flatten()
                index = v1.argsort()[-count:][::-1]
                fo_maxs = v1[index]
                result.append(fo_maxs)
        result = np.array(result)
        return result

def ob_fo_std(ob,fo):
    Fo_shape = fo.shape
    Ob_shape = ob.shape

    Ob_shpe_list = list(Ob_shape)
    size = len(Ob_shpe_list)
    ind = -size
    Fo_Ob_index = list(Fo_shape[ind:])
    if Fo_Ob_index != Ob_shpe_list:
        print('预报数据和观测数据维度不匹配')
        return

    if len(fo.shape) == len(ob.shape):
        result = [np.std(ob),np.std(fo)]
    else:
        result = [np.std(ob)]
        for i in range(Fo_shape[0]):
            result.append(np.std(fo[i,:]))
    result = np.array(result)
    return result

def ob_fo_cv(ob,fo):
    '''
    观测和预报的离差
    :param ob:
    :param fo:
    :return:
    '''
    Fo_shape = fo.shape
    Ob_shape = ob.shape

    Ob_shpe_list = list(Ob_shape)
    size = len(Ob_shpe_list)
    ind = -size
    Fo_Ob_index = list(Fo_shape[ind:])
    if Fo_Ob_index != Ob_shpe_list:
        print('预报数据和观测数据维度不匹配')
        return

    if len(fo.shape) == len(ob.shape):
        result = [np.std(ob)/np.mean(ob),np.std(fo)/np.mean(fo)]
    else:
        result = [np.std(ob)/np.mean(ob)]
        for i in range(Fo_shape[0]):
            result.append(np.std(fo[i,:])/np.mean(fo[i,:]))
    result = np.array(result)
    return result


def cscs(Ob,Fo):
    '''
       :param ob: 观测降水序列
       :param fo: 预报降水序列
       :return: 观测和预报各自的平均降水强度计算相关的中间量
       '''
    cscs_list = []
    Fo_shape = Fo.shape
    Ob_shape = Ob.shape

    Ob_shpe_list = list(Ob_shape)
    size = len(Ob_shpe_list)
    ind = -size
    Fo_Ob_index = list(Fo_shape[ind:])
    if Fo_Ob_index != Ob_shpe_list:
        print('预报数据和观测数据维度不匹配')
        return
    Ob_shpe_list.insert(0, -1)
    new_Fo_shape = tuple(Ob_shpe_list)
    new_Fo = Fo.reshape(new_Fo_shape)
    new_Fo_shape = new_Fo.shape

    ob_not_0 = Ob[Ob >= 0.1]
    cob = ob_not_0.size
    sob = np.sum(ob_not_0)

    for line in range(new_Fo_shape[0]):
        foi = new_Fo[line, :]
        fo_not_0 = foi[foi >= 0.1]
        cfo = fo_not_0.size
        sfo = np.sum(fo_not_0)
        cscs_list.append(np.array([cob,sob, cfo,sfo]))

    cscs_np = np.array(cscs_list)
    shape = list(Fo_shape[:ind])
    shape.append(4)

    cscs_array =cscs_np.reshape(shape)
    return cscs_array

def ob_fo_precipitation_strength_cscs(cscs_array):

    if len(cscs_array.shape) == 1:
        cscs_array = cscs_array.reshape(1,1, 4)
    cob = cscs_array[..., 0]
    sob = cscs_array[..., 1]
    cfo = cscs_array[..., 2]
    sfo = cscs_array[..., 3]

    sob[cob==0] = IV
    cob[cob == 0] = 1
    sfo[cfo == 0] =IV
    cfo[cfo == 0] = 1

    shape1 = list(cscs_array.shape)
    if len(cscs_array.shape) == 2:
        result = np.zeros((2, shape1[0]))
        result[0, :] = sob / cob
        result[1, :] = sfo / cfo
    else:
        result = np.zeros((1 + shape1[0], shape1[1]))
        result[0, :] = sob / cob
        result[1, :] = sfo / cfo

    return result


def ob_fo_precipitation_strength(ob,fo):
    '''
    :param ob: 观测降水序列
    :param fo: 预报降水序列
    :return: 观测和预报各自的平均降水强度，平均降水强度等于降水量大于等于0.1mm的站次的降水的平均值
    '''
    Fo_shape = fo.shape
    Ob_shape = ob.shape

    Ob_shpe_list = list(Ob_shape)
    size = len(Ob_shpe_list)
    ind = -size
    Fo_Ob_index = list(Fo_shape[ind:])
    if Fo_Ob_index != Ob_shpe_list:
        print('预报数据和观测数据维度不匹配')
        return

    ob_not_0 = ob[ob >= 0.1]
    result = [np.mean(ob_not_0)]
    if len(fo.shape) == len(ob.shape):
        fo_not_0 = fo[fo>=0.1]
        result.append(np.mean(fo_not_0))
    else:
        for i in range(Fo_shape[0]):
            foi= fo[i,:]
            fo_not_0 = foi[foi >= 0.1]
            result.append(np.mean(fo_not_0))
    result = np.array(result)
    return result


def ob_fo_precipitation_strenght(ob,fo):
    result = ob_fo_precipitation_strength(ob,fo)
    return result



def ob_mean(Ob, Fo=None):
    '''
    计算观测样本的平均
    -----------------------------
    :param Ob: 实况数据  任意维numpy数组
    :param Fo: None或任意数据，它的存在是为了使得参数规范化，方便更高级的封装
    :return: 实数
    '''
    return np.mean(Ob)


def fo_mean(Ob, Fo):
    '''
    计算观测样本的平均
    -----------------------------
    :param Ob: None或任意数据，它的存在是为了使得参数规范化，方便更高级的封装
    :param Fo: 预报数据  任意维numpy数组
    :return: 实数
    '''
    Fo_mean_list = []
    Fo_shape = Fo.shape
    Ob_shape = Ob.shape

    Ob_shpe_list = list(Ob_shape)
    size = len(Ob_shpe_list)
    ind = -size
    Fo_Ob_index = list(Fo_shape[ind:])
    if Fo_Ob_index != Ob_shpe_list:
        print('预报数据和观测数据维度不匹配')
        return
    if len(Fo_shape)> len(Ob_shape):
        Ob_shpe_list.insert(0, -1)
        new_Fo_shape = tuple(Ob_shpe_list)
        new_Fo = Fo.reshape(new_Fo_shape)
        new_Fo_shape = new_Fo.shape
        for line in range(new_Fo_shape[0]):
            Fo_mean = np.mean(new_Fo[line, :])
            Fo_mean_list.append(Fo_mean)
        Fo_mean_array = np.array(Fo_mean_list)
        shape = list(Fo_shape[:ind])
        Fo_mean_array = Fo_mean_array.reshape(shape)
    else:
        Fo_mean_array = np.mean(Fo)
    return Fo_mean_array


def tc_count(Ob, Fo, grade_list = [2]):
    '''
    计算准确率的中间结果
    :param Ob:
    :param Fo:
    :param threshold:
    :return:
    '''
    if not isinstance(grade_list,list):
        grade_list = [grade_list]
    correct_rate_list = []
    Fo_shape = Fo.shape
    Ob_shape = Ob.shape

    Ob_shpe_list = list(Ob_shape)
    size = len(Ob_shpe_list)
    ind = -size
    Fo_Ob_index = list(Fo_shape[ind:])
    if Fo_Ob_index != Ob_shpe_list:
        print('预报数据和观测数据维度不匹配')
        return
    Ob_shpe_list.insert(0, -1)
    new_Fo_shape = tuple(Ob_shpe_list)
    new_Fo = Fo.reshape(new_Fo_shape)
    new_Fo_shape = new_Fo.shape
    for line in range(new_Fo_shape[0]):
        total_count = Ob.size
        error = np.abs(new_Fo[line, :] - Ob)
        count_list = [total_count]
        for grade in grade_list:
            index = np.where(error <= grade)
            count_list.append(len(index[0]))
        correct_rate_list.append(count_list)
    correct_rate_np = np.array(correct_rate_list)
    shape = list(Fo_shape[:ind])
    shape.append(1 + len(grade_list))
    correct_rate_array = correct_rate_np.reshape(shape)
    return correct_rate_array


def correct_rate(Ob, Fo, grade_list = [2]):
    '''
    计算准确率
    :param Ob:
    :param Fo:
    :param threshold:
    :return:
    '''

    tc_array = tc_count(Ob, Fo, grade_list)
    crate = correct_rate_tc(tc_array)
    return crate

def wrong_rate(Ob,Fo,grade_list = [2],unit = 1):
    '''
    计算错误率
    :param Ob:
    :param Fo:
    :param threshold:
    :return:
    '''

    tc_array = tc_count(Ob, Fo, grade_list)
    wrate = wrong_rate_tc(tc_array,unit = unit)

    return wrate


def wrong_rate_tc(tc_count_array,unit = 1):
    '''
    计算错误率
    :param Ob:
    :param Fo:
    :param threshold:
    :return:
    '''
    crate = correct_rate_tc(tc_count_array)
    wrate = 1-crate
    if unit =="%":
        wrate *=100
    return wrate

def correct_rate_tc(tc_count_array):
    '''
    :param tc_count_array:
    :return:
    '''
    if tc_count_array.shape[-1] > 2:
        total_count =  tc_count_array[..., 0]
        total_count = total_count.reshape((-1,1))
        cr1 = tc_count_array[..., 1:] / total_count
    else:
        cr1 = tc_count_array[..., 1] / tc_count_array[..., 0]

    return cr1


def tlfo(Ob,Fo):

    '''
    计算RMSF的中间结果
    -----------------------------
    :param Ob: 实况数据  任意维numpy数组
    :param Fo: 预测数据 任意维numpy数组,Fo.shape 和Ob.shape一致
    :return: 一维numpy数组，其内容依次为总样本数、log(fo/ob)^2 的总和
    '''

    tlfo_list = []
    Fo_shape = Fo.shape
    Ob_shape = Ob.shape

    Ob_shpe_list = list(Ob_shape)
    size = len(Ob_shpe_list)
    ind = -size
    Fo_Ob_index = list(Fo_shape[ind:])
    if Fo_Ob_index != Ob_shpe_list:
        print('预报数据和观测数据维度不匹配')
        return
    Ob_shpe_list.insert(0, -1)
    new_Fo_shape = tuple(Ob_shpe_list)
    new_Fo = Fo.reshape(new_Fo_shape)
    new_Fo_shape = new_Fo.shape
    for line in range(new_Fo_shape[0]):
        fo_ob = np.array([new_Fo[line, :],Ob])
        min_ob_fo = np.min(fo_ob,axis=0)
        max_ob_fo = np.max(fo_ob,axis=0)
        index = np.where((min_ob_fo>=0.1)|(max_ob_fo>=1.0))
        ob_s = Ob[index]
        fo_s = new_Fo[line,:][index]
        ob_s[ob_s<0.1] = 0.1
        fo_s[fo_s<0.1] = 0.1
        total_count = ob_s.size
        e_sum = np.sum(np.power(np.log(fo_s/ob_s),2))
        tlfo_list.append(np.array([total_count, e_sum]))
    tlfo_np = np.array(tlfo_list)
    shape = list(Fo_shape[:ind])
    shape.append(2)

    tlfo_array = tlfo_np.reshape(shape)
    return tlfo_array

def rmsf(Ob,Fo):
    '''

    :param Ob:
    :param Fo:
    :return:
    '''
    tlfo_array = tlfo(Ob,Fo)
    return rmsf_tlfo(tlfo_array)

def rmsf_tlfo(tlfo_array):
    mean_log2 = tlfo_array[..., 1] / tlfo_array[..., 0]
    rmsf = np.exp(np.sqrt(mean_log2))
    return rmsf


def tase(Ob, Fo,weight = None):
    '''
    计算平均误差、平均绝对误差、均方误差、均方根误差的中间结果
    -----------------------------
    :param Ob: 实况数据  任意维numpy数组
    :param Fo: 预测数据 任意维numpy数组,Fo.shape 和Ob.shape一致
    :return: 一维numpy数组，其内容依次为总样本数、误差总和、绝对误差总和、误差平方总和
    '''

    tase_list = []
    Fo_shape = Fo.shape
    Ob_shape = Ob.shape
    Ob_shpe_list = list(Ob_shape)
    size = len(Ob_shpe_list)
    ind = -size
    Fo_Ob_index = list(Fo_shape[ind:])
    if Fo_Ob_index != Ob_shpe_list:
        print('预报数据和观测数据维度不匹配')
        return
    Ob_shpe_list.insert(0, -1)
    new_Fo_shape = tuple(Ob_shpe_list)
    new_Fo = Fo.reshape(new_Fo_shape)
    new_Fo_shape = new_Fo.shape
    if weight is None:
        for line in range(new_Fo_shape[0]):
            total_count = Ob.size
            e_sum = np.sum(new_Fo[line, :] - Ob)
            ae_sum = np.sum(np.abs(new_Fo[line, :] - Ob))
            se_sum = np.sum(np.square(new_Fo[line, :] - Ob))
            tase_list.append(np.array([total_count, e_sum, ae_sum, se_sum]))
    else:
        for line in range(new_Fo_shape[0]):
            total_count = np.sum(weight)
            e_sum = np.sum((new_Fo[line, :] - Ob) * weight)
            ae_sum = np.sum(np.abs(new_Fo[line, :] - Ob) * weight)
            se_sum = np.sum(np.square(new_Fo[line, :] - Ob)*weight)
            tase_list.append(np.array([total_count, e_sum, ae_sum, se_sum]))

    tase_np = np.array(tase_list)
    shape = list(Fo_shape[:ind])
    shape.append(4)
    tase_array = tase_np.reshape(shape)
    return tase_array


def max_error(Ob,Fo):
    me_list = []
    Fo_shape = Fo.shape
    Ob_shape = Ob.shape

    Ob_shpe_list = list(Ob_shape)
    size = len(Ob_shpe_list)
    ind = -size
    Fo_Ob_index = list(Fo_shape[ind:])
    if Fo_Ob_index != Ob_shpe_list:
        print('预报数据和观测数据维度不匹配')
        return
    if len(Fo_shape)> len(Ob_shape):
        Ob_shpe_list.insert(0, -1)
        new_Fo_shape = tuple(Ob_shpe_list)
        new_Fo = Fo.reshape(new_Fo_shape)
        new_Fo_shape = new_Fo.shape
        for line in range(new_Fo_shape[0]):
            error = np.max(new_Fo[line, :] - Ob)
            me_list.append(error)
        error_array = np.array(me_list)
        shape = list(Fo_shape[:ind])
        error_array = error_array.reshape(shape)
    else:
        error_array  = np.max(Fo - Ob)
    return error_array

def min_error(Ob,Fo):
    me_list = []
    Fo_shape = Fo.shape
    Ob_shape = Ob.shape

    Ob_shpe_list = list(Ob_shape)
    size = len(Ob_shpe_list)
    ind = -size
    Fo_Ob_index = list(Fo_shape[ind:])
    if Fo_Ob_index != Ob_shpe_list:
        print('预报数据和观测数据维度不匹配')
        return
    if len(Fo_shape)> len(Ob_shape):
        Ob_shpe_list.insert(0, -1)
        new_Fo_shape = tuple(Ob_shpe_list)
        new_Fo = Fo.reshape(new_Fo_shape)
        new_Fo_shape = new_Fo.shape
        for line in range(new_Fo_shape[0]):
            error = np.min(new_Fo[line, :] - Ob)
            me_list.append(error)
        error_array = np.array(me_list)
        shape = list(Fo_shape[:ind])
        error_array = error_array.reshape(shape)
    else:
        error_array  = np.min(Fo - Ob)
    return error_array


def max_abs_error(Ob,Fo):
    me_list = []
    Fo_shape = Fo.shape
    Ob_shape = Ob.shape

    Ob_shpe_list = list(Ob_shape)
    size = len(Ob_shpe_list)
    ind = -size
    Fo_Ob_index = list(Fo_shape[ind:])
    if Fo_Ob_index != Ob_shpe_list:
        print('预报数据和观测数据维度不匹配')
        return
    if len(Fo_shape)> len(Ob_shape):
        Ob_shpe_list.insert(0, -1)
        new_Fo_shape = tuple(Ob_shpe_list)
        new_Fo = Fo.reshape(new_Fo_shape)
        new_Fo_shape = new_Fo.shape
        for line in range(new_Fo_shape[0]):
            error = np.max(np.abs(new_Fo[line, :] - Ob))
            me_list.append(error)
        error_array = np.array(me_list)
        shape = list(Fo_shape[:ind])
        error_array = error_array.reshape(shape)
    else:
        error_array  = np.max(np.abs(Fo - Ob))
    return error_array


def me(Ob, Fo,weight = None):
    '''
    me 求两组数据的误差平均值
    -----------------------------
    :param Ob: 实况数据  任意维numpy数组
    :param Fo: 预测数据 任意维numpy数组,Fo.shape 和Ob.shape一致
    :return: 负无穷到正无穷的实数，最优值为0
    '''

    if weight is None:
        me_list = []
        Fo_shape = Fo.shape
        Ob_shape = Ob.shape

        Ob_shpe_list = list(Ob_shape)
        size = len(Ob_shpe_list)
        ind = -size
        Fo_Ob_index = list(Fo_shape[ind:])
        if Fo_Ob_index != Ob_shpe_list:
            print('预报数据和观测数据维度不匹配')
            return
        if len(Fo_shape)> len(Ob_shape):
            Ob_shpe_list.insert(0, -1)
            new_Fo_shape = tuple(Ob_shpe_list)
            new_Fo = Fo.reshape(new_Fo_shape)
            new_Fo_shape = new_Fo.shape
            for line in range(new_Fo_shape[0]):
                mean_error = np.mean(new_Fo[line, :] - Ob)
                me_list.append(mean_error)
            mean_error_array = np.array(me_list)
            shape = list(Fo_shape[:ind])
            mean_error_array = mean_error_array.reshape(shape)
        else:
            mean_error_array  = np.mean(Fo - Ob)
    else:
        tase_array = tase(Ob,Fo,weight)
        mean_error_array = me_tase(tase_array)

    return mean_error_array


def me_tase(tase_array):
    '''
    me 求两组数据的误差平均值
    :param tase_array:包含命中空报和漏报的多维数组，其中最后一维长度为4，分别记录了（total_count,e_sum,ae_sum,se_sum）
    （样本数，误差和、绝对误差和，误差平方和），它由tase返回
    :return: 负无穷到正无穷的实数，最优值为0
    '''
    mean_error = tase_array[..., 1] / tase_array[..., 0]
    return mean_error


def mae(Ob, Fo,weight = None):
    '''
    me 求两组数据的误差平均值
    -----------------------------
    :param Ob: 实况数据  任意维numpy数组
    :param Fo: 预测数据 任意维numpy数组,Fo.shape 和Ob.shape一致
    :return: 负无穷到正无穷的实数，最优值为0
    '''
    if weight is None:
        mae_list = []
        Fo_shape = Fo.shape
        Ob_shape = Ob.shape

        Ob_shpe_list = list(Ob_shape)
        size = len(Ob_shpe_list)
        ind = -size
        Fo_Ob_index = list(Fo_shape[ind:])
        if Fo_Ob_index != Ob_shpe_list:
            print('预报数据和观测数据维度不匹配')
            return
        if len(Fo_shape) == len(Ob_shape):
            mean_abs_error = np.mean(np.abs(Fo - Ob))
            return mean_abs_error
        else:
            Ob_shpe_list.insert(0, -1)
            new_Fo_shape = tuple(Ob_shpe_list)
            new_Fo = Fo.reshape(new_Fo_shape)
            new_Fo_shape = new_Fo.shape
            for line in range(new_Fo_shape[0]):
                mean_abs_error = np.mean(np.abs(new_Fo[line, :] - Ob))
                mae_list.append(mean_abs_error)
            mean_error_array = np.array(mae_list)
            shape = list(Fo_shape[:ind])
            mean_abs_error_array = mean_error_array.reshape(shape)
            return mean_abs_error_array
    else:
        tase_array = tase(Ob, Fo, weight)
        mean_abs_error_array = mae_tase(tase_array)

        return mean_abs_error_array


def mae_tase(tase_array):
    '''
    mean_abs_error,求两组数据的平均绝对误差
    :param tase_array:包含命中空报和漏报的多维数组，其中最后一维长度为4，分别记录了（total_count,e_sum,ae_sum,se_sum）
    （样本数，误差和、绝对误差和，误差平方和），它由tase返回
    :return: 0到无穷大，最优值为0
    '''
    mean_abs_error = tase_array[..., 2] / tase_array[..., 0]
    return mean_abs_error


def mse(Ob, Fo,weight = None):
    '''
    mean_sqrt_error, 求两组数据的均方误差
    ----------------------------------
    :param Ob: 实况数据  任意维numpy数组
    :param Fo: 预测数据 任意维numpy数组,Fo.shape 和Ob.shape一致
    :return: 0到无穷大，最优值为0
    '''
    if weight is None:
        mse_list = []
        Fo_shape = Fo.shape
        Ob_shape = Ob.shape

        Ob_shpe_list = list(Ob_shape)
        size = len(Ob_shpe_list)
        ind = -size
        Fo_Ob_index = list(Fo_shape[ind:])
        if Fo_Ob_index != Ob_shpe_list:
            print('预报数据和观测数据维度不匹配')
            return
        if len(Fo_shape) == len(Ob_shape):
            mean_square_error = np.mean(np.square(Fo - Ob))
            return mean_square_error
        else:
            Ob_shpe_list.insert(0, -1)
            new_Fo_shape = tuple(Ob_shpe_list)
            new_Fo = Fo.reshape(new_Fo_shape)
            new_Fo_shape = new_Fo.shape
            for line in range(new_Fo_shape[0]):
                mean_square_error = np.mean(np.square(new_Fo[line, :] - Ob))
                mse_list.append(mean_square_error)
            mean_sqrt_array = np.array(mse_list)
            shape = list(Fo_shape[:ind])
            mean_sqrt_error_array = mean_sqrt_array.reshape(shape)
            return mean_sqrt_error_array
    else:
        tase_array = tase(Ob, Fo, weight)
        mean_squre_error = mse_tase(tase_array)

        return mean_squre_error

def mse_tase(tase_array):
    '''
    mse 求两组数据的均方误差
    :param tase_array:包含命中空报和漏报的多维数组，其中最后一维长度为4，分别记录了（total_count,e_sum,ae_sum,se_sum）
    （样本数，误差和、绝对误差和，误差平方和），它由tase返回
    :return: 0到无穷大，最优值为0
    '''
    mean_squre_error = tase_array[..., 3] / tase_array[..., 0]
    return mean_squre_error


def rmse(Ob, Fo,weight = None):
    '''
    root_mean_square_error 求两组数据的均方根误差
    ------------------------------
    :param Ob: 实况数据  任意维numpy数组
    :param Fo: 预测数据 任意维numpy数组,Fo.shape 和Ob.shape一致
    :return: 0到无穷大，最优值为0
    '''
    if weight is None:
        rmse_list = []
        Fo_shape = Fo.shape
        Ob_shape = Ob.shape

        Ob_shpe_list = list(Ob_shape)
        size = len(Ob_shpe_list)
        ind = -size
        Fo_Ob_index = list(Fo_shape[ind:])
        if Fo_Ob_index != Ob_shpe_list:
            print('预报数据和观测数据维度不匹配')
            return
        if len(Fo_shape) == len(Ob_shape):
            mean_square_error = np.sqrt(np.mean(np.square(Fo - Ob)))
            return mean_square_error
        else:
            Ob_shpe_list.insert(0, -1)
            new_Fo_shape = tuple(Ob_shpe_list)
            new_Fo = Fo.reshape(new_Fo_shape)
            new_Fo_shape = new_Fo.shape
            for line in range(new_Fo_shape[0]):
                root_mean_sqrt_error = np.sqrt(np.mean(np.square(new_Fo[line, :] - Ob)))
                rmse_list.append(root_mean_sqrt_error)
            root_mean_sqrt_array = np.array(rmse_list)
            shape = list(Fo_shape[:ind])
            root_mean_sqrt_error_array = root_mean_sqrt_array.reshape(shape)
            return root_mean_sqrt_error_array
    else:
        tase_array = tase(Ob, Fo, weight)
        root_mean_sqrt_error = rmse_tase(tase_array)

        return root_mean_sqrt_error

def rmse_tase(tase_array):
    '''
    mse 求两组数据的均方根误差
    :param tase_array:包含命中空报和漏报的多维数组，其中最后一维长度为4，分别记录了（total_count,e_sum,ae_sum,se_sum）
    （样本数，误差和、绝对误差和，误差平方和），它由tase返回
    :return: 0到无穷大，最优值为0
    '''
    root_mean_sqrt_error = np.sqrt(tase_array[..., 3] / tase_array[..., 0])
    return root_mean_sqrt_error

def si(Ob,Fo):
    '''

    :param Ob:
    :param Fo:
    :return: 均方根误差/观测平均
    '''
    si_list = []
    Fo_shape = Fo.shape
    Ob_shape = Ob.shape

    Ob_shpe_list = list(Ob_shape)
    size = len(Ob_shpe_list)
    ind = -size
    Fo_Ob_index = list(Fo_shape[ind:])
    mean_ob = np.mean(Ob)
    if Fo_Ob_index != Ob_shpe_list:
        print('预报数据和观测数据维度不匹配')
        return
    if len(Fo_shape) == len(Ob_shape):
        root_mean_sqrt_error = np.sqrt(np.mean(np.square(Fo - Ob)))

        si = root_mean_sqrt_error/mean_ob
        return si
    else:
        Ob_shpe_list.insert(0, -1)
        new_Fo_shape = tuple(Ob_shpe_list)
        new_Fo = Fo.reshape(new_Fo_shape)
        new_Fo_shape = new_Fo.shape
        for line in range(new_Fo_shape[0]):
            root_mean_sqrt_error = np.sqrt(np.mean(np.square(new_Fo[line, :] - Ob)))
            si = root_mean_sqrt_error/mean_ob
            si_list.append(si)
        si_array = np.array(si_list)
        shape = list(Fo_shape[:ind])
        si_array = si_array.reshape(shape)
        return si_array


def bias_m(Ob, Fo):
    '''
    均值偏差 求预测数据和实况数据的平均值的比
    ------------------------------
    :param Ob: 实况数据  任意维numpy数组
    :param Fo: 预测数据 任意维numpy数组,Fo.shape 和Ob.shape一致
    :return:  0到正无穷，最优值为1
    '''
    mean_ob = np.mean(Ob)
    if mean_ob == 0:
        bias0 = IV
    else:
        bias_m_list = []
        Fo_shape = Fo.shape
        Ob_shape = Ob.shape

        Ob_shpe_list = list(Ob_shape)
        size = len(Ob_shpe_list)
        ind = -size
        Fo_Ob_index = list(Fo_shape[ind:])
        if Fo_Ob_index != Ob_shpe_list:
            print('预报数据和观测数据维度不匹配')
            return
        if len(Fo_shape) == len(Ob_shape):
            bias0 = np.mean(Fo) / mean_ob
        else:
            Ob_shpe_list.insert(0, -1)
            new_Fo_shape = tuple(Ob_shpe_list)
            new_Fo = Fo.reshape(new_Fo_shape)
            new_Fo_shape = new_Fo.shape
            for line in range(new_Fo_shape[0]):
                bias_piece = np.mean(new_Fo[line, :]) / mean_ob
                bias_m_list.append(bias_piece)
            bias_m_np = np.array(bias_m_list)
            shape = list(Fo_shape[:ind])
            bias0 = bias_m_np.reshape(shape)
    return bias0

def bias_tmmsss(tmmsss_array):
    '''
    均值偏差 求预测数据和实况数据的平均值的比
    :param tmmsss_array: 包含命中空报和漏报的多维数组，其中最后一维长度为6，分别记录了（count,mx,my,sxx,syy,sxy）
    :return:
    '''
    mean_ob = tmmsss_array[..., 1] + 0
    mean_fo = tmmsss_array[..., 2]
    if mean_ob.size == 1:
        if mean_ob == 0:
            bias0 = IV
        else:
            bias0 = mean_fo / mean_ob
    else:
        mean_ob[mean_ob == 0] = IV
        bias0 = mean_ob / mean_fo
        bias0[mean_ob == IV] = IV
    return bias0


def corr(Ob, Fo,weight = None):
    '''
    相关系数，求实况数据还和预测数据之间的相关系数
    -----------------------------
    :param Ob: 实况数据  任意维numpy数组
    :param Fo: 预测数据 任意维numpy数组,Fo.shape 和Ob.shape一致
    :return: corr0
    '''
    tmmsss_array = tmmsss(Ob,Fo,weight = weight)
    corr0 = corr_tmmsss(tmmsss_array)
    return corr0

def corr_rank(Ob,Fo):
    rcc_list = []
    Fo_shape = Fo.shape
    Ob_shape = Ob.shape

    Ob_shpe_list = list(Ob_shape)
    size = len(Ob_shpe_list)
    ind = -size
    Fo_Ob_index = list(Fo_shape[ind:])

    if Fo_Ob_index != Ob_shpe_list:
        print('预报数据和观测数据维度不匹配')
        return
    if len(Fo_shape) == len(Ob_shape):

        r_ob = np.argsort(np.argsort(Ob.flatten()))
        r_fo = np.argsort(np.argsort(Fo.flatten()))
        n = r_ob.size
        rcc = 1 - 6 * np.sum(np.power(r_fo-r_ob,2))/(n * (n*n-1))

        #rcc = 12*np.sum(r_fo * r_fo)/(n * (n*n-1)) - 3*(n+1)/(n-1)
        return rcc
    else:
        Ob_shpe_list.insert(0, -1)
        new_Fo_shape = tuple(Ob_shpe_list)
        new_Fo = Fo.reshape(new_Fo_shape)
        new_Fo_shape = new_Fo.shape
        r_ob = np.argsort(np.argsort(Ob.flatten()))
        n = r_ob.size
        for line in range(new_Fo_shape[0]):
            r_fo = np.argsort(np.argsort(new_Fo[line, :].flatten()))
            rcc = 1 - 6 * np.sum(np.power(r_fo-r_ob,2))/(n * (n*n-1))
            rcc_list.append(rcc)
        rcc_array = np.array(rcc_list)
        shape = list(Fo_shape[:ind])
        rcc_array = rcc_array.reshape(shape)
        return rcc_array


def residual_error(Ob,Fo):
    '''
    线性回归的残差， 它等于残差率 *  观测数据的方差
    :param Ob:
    :param Fo:
    :return:
    '''
    tmmsss_array = tmmsss(Ob, Fo)
    re = residual_error_tmmsss(tmmsss_array)
    return re

def residual_error_tmmsss(tmmsss_array):
    '''
    线性回归的残差， 它等于残差率 *  观测数据的方差
    :param tmmsss_array:
    :return:
    '''
    rer = residual_error_rate_tmmsss(tmmsss_array)
    sxx = tmmsss_array[..., 3]
    re = rer * np.sqrt(sxx)
    return re

def residual_error_rate(Ob, Fo):
    '''
    线性回归的残差率，等于 1 - corr * corr
    -----------------------------
    :param Ob: 实况数据  任意维numpy数组
    :param Fo: 预测数据 任意维numpy数组,Fo.shape 和Ob.shape一致
    :return: corr0
    '''
    tmmsss_array = tmmsss(Ob,Fo)
    rer = residual_error_rate_tmmsss(tmmsss_array)
    return rer

def residual_error_rate_tmmsss(tmmsss_array):
    '''
    线性回归的残差率，等于 1 - corr * corr
    :param tmmsss_array:
    :return:
    '''
    corr0 = corr_tmmsss(tmmsss_array)
    rer = np.sqrt(1 - np.power(corr0, 2))
    return rer


def corr_tmmsss(tmmsss_array):
    '''
    相关系数，求实况数据还和预测数据之间的相关系数
    :param tmmsss_array: 包含命中空报和漏报的多维数组，其中最后一维长度为6，分别记录了（count,mx,my,sxx,syy,sxy）
    :return:
    '''
    sxx = tmmsss_array[..., 3]
    syy = tmmsss_array[..., 4]
    sxy = tmmsss_array[..., 5]
    sxxsyy = np.sqrt(sxx * syy)
    if sxxsyy.size == 1:
        if sxxsyy == 0:
            sxxsyy = 1e-10
    else:
        sxxsyy[sxxsyy == 0] = 1e-10
    corr = sxy / sxxsyy
    return corr

def ob_fo_sum_tmmsss(tmmsss_array):
    '''
    相关系数，求实况数据还和预测数据之间的相关系数
    :param tmmsss_array: 包含命中空报和漏报的多维数组，其中最后一维长度为6，分别记录了（count,mx,my,sxx,syy,sxy）
    :return:
    '''
    if len(tmmsss_array.shape) == 1:
        tmmsss_array = tmmsss_array.reshape(1, 1, 6)
    mx = tmmsss_array[..., 1]
    my = tmmsss_array[..., 2]
    shape1 = list(tmmsss_array.shape)
    if len(tmmsss_array.shape) == 2:
        result = np.zeros((2, shape1[0]))
        result[0, :] = mx * tmmsss_array[..., 0]
        result[1, :] = my * tmmsss_array[..., 0]
    else:
        result = np.zeros((1 + shape1[0], shape1[1]))
        result[0, :] = mx * tmmsss_array[..., 0]
        result[1:, :] = my * tmmsss_array[..., 0]
    return result



def ob_fo_mean_tmmsss(tmmsss_array):
    '''
    相关系数，求实况数据还和预测数据之间的相关系数
    :param tmmsss_array: 包含命中空报和漏报的多维数组，其中最后一维长度为6，分别记录了（count,mx,my,sxx,syy,sxy）
    :return:
    '''

    if len(tmmsss_array.shape) == 1:
        tmmsss_array = tmmsss_array.reshape(1, 1, 6)
    mx = tmmsss_array[..., 1]
    my = tmmsss_array[..., 2]
    shape1 = list(tmmsss_array.shape)
    if len(tmmsss_array.shape) == 2:
        result = np.zeros((2, shape1[0]))
        result[0, :] = mx
        result[1, :] = my
    else:
        result = np.zeros((1 + shape1[0], shape1[1]))
        result[0, :] = mx
        result[1:, :] = my
    return result



def ob_fo_std_tmmsss(tmmsss_array):
    '''
    相关系数，求实况数据还和预测数据之间的相关系数
    :param tmmsss_array: 包含命中空报和漏报的多维数组，其中最后一维长度为6，分别记录了（count,mx,my,sxx,syy,sxy）
    :return:
    '''

    if len(tmmsss_array.shape) == 1:
        tmmsss_array = tmmsss_array.reshape(1, 1, 6)
    sx = tmmsss_array[..., 3]
    sy = tmmsss_array[..., 4]
    shape1 = list(tmmsss_array.shape)
    if len(tmmsss_array.shape) == 2:
        result = np.zeros((2, shape1[0]))
        result[0, :] = sx
        result[1, :] = sy
    else:
        result = np.zeros((1 + shape1[0], shape1[1]))
        result[0, :] = sx
        result[1:, :] = sy
    result = np.sqrt(result)
    return result

def tmmsss(Ob, Fo,weight = None):
    '''
    统计相关系数等检验量所需的中间变量
    :param Ob: 实况数据  任意维numpy数组
    :param Fo: 预测数据 任意维numpy数组,Fo.shape 和Ob.shape一致
    :return: numpy 一维数组，其元素为根据Ob和Fo
    计算出的（样本数，观测平均值，预报平均值，观测方差，预报方差，协方差
    '''
    tmmsss_array_list = []
    Fo_shape = Fo.shape
    Ob_shape = Ob.shape

    Ob_shpe_list = list(Ob_shape)
    size = len(Ob_shpe_list)
    ind = -size
    Fo_Ob_index = list(Fo_shape[ind:])
    if Fo_Ob_index != Ob_shpe_list:
        print('预报数据和观测数据维度不匹配')
        return
    Ob_shpe_list.insert(0, -1)
    new_Fo_shape = tuple(Ob_shpe_list)
    new_Fo = Fo.reshape(new_Fo_shape)
    new_Fo_shape = new_Fo.shape
    ob_f = Ob.flatten()
    if weight is None:
        count = Ob.size
        mx = np.mean(ob_f)
        for line in range(new_Fo_shape[0]):
            fo_f = new_Fo[line, :].flatten()
            my = np.mean(fo_f)
            dx = ob_f - mx
            dy = fo_f - my
            sxx = np.mean(np.power(dx, 2))
            syy = np.mean(np.power(dy, 2))
            sxy = np.mean(dx * dy)
            tmmsss_array_list.append(np.array([count, mx, my, sxx, syy, sxy]))
    else:
        weight_f = weight.flatten()
        count = np.sum(weight_f)
        mx = np.sum(ob_f * weight_f) /count
        for line in range(new_Fo_shape[0]):
            fo_f = new_Fo[line, :].flatten()
            my = np.sum(fo_f * weight_f) /count
            dx = ob_f - mx
            dy = fo_f - my
            sxx = np.sum(np.power(dx, 2) * weight_f)/count
            syy = np.sum(np.power(dy, 2) * weight_f)/count
            sxy = np.sum(dx * dy * weight_f) / count
            tmmsss_array_list.append(np.array([count, mx, my, sxx, syy, sxy]))
    tmmsss_array = np.array(tmmsss_array_list)
    shape = list(Fo_shape[:ind])
    shape.append(6)
    tmmsss_array = tmmsss_array.reshape(shape)
    return tmmsss_array


def tmmsss_merge(tmmsss0, tmmsss1):
    '''
    将两份包含样本数、平均值和方差、协方差的中间结果合并
    :param tmmsss0: 长度6的一维数组，分别记录了（count,mx,my,sxx,syy,sxy）
    :param tmmsss1: 长度6的一维数组，分别记录了（count,mx,my,sxx,syy,sxy）
    :return: 长度6的一维数组，分别记录了（count,mx,my,sxx,syy,sxy）
    '''
    if np.isnan(tmmsss0[0]).any():
        return tmmsss1
    elif np.isnan(tmmsss1[0]).any():
        return tmmsss0
    tmmsss_array_list = []
    tmmsss0_shape = list(tmmsss0.shape)
    tmmsss1_shape = list(tmmsss1.shape)
    if tmmsss0_shape != tmmsss1_shape:
        print('tmmsss0和tmmsss1维度不匹配')
        return
    tmmsss0 = tmmsss0.reshape((-1, 6))
    tmmsss1 = tmmsss1.reshape((-1, 6))
    new_tmmsss1_shape = tmmsss1.shape
    for line in range(new_tmmsss1_shape[0]):
        tmmsss1_piece = tmmsss1[line, :]
        tmmsss0_piece = tmmsss0[line, :]
        count_0 = tmmsss0_piece[0]
        mx_0 = tmmsss0_piece[1]
        my_0 = tmmsss0_piece[2]
        sxx_0 = tmmsss0_piece[3]
        syy_0 = tmmsss0_piece[4]
        sxy_0 = tmmsss0_piece[5]
        count_1 = tmmsss1_piece[0]
        mx_1 = tmmsss1_piece[1]
        my_1 = tmmsss1_piece[2]
        sxx_1 = tmmsss1_piece[3]
        syy_1 = tmmsss1_piece[4]
        sxy_1 = tmmsss1_piece[5]
        _, _, sxx_total = ss_iteration(count_0, mx_0, sxx_0, count_1, mx_1, sxx_1)
        _, _, syy_total = ss_iteration(count_0, my_0, syy_0, count_1, my_1, syy_1)
        count_total, mx_total, my_total, sxy_total = sxy_iteration(count_0, mx_0, my_0, sxy_0,
                                                                   count_1, mx_1, my_1, sxy_1)

        tmmsss_array_list.append(np.array([count_total, mx_total, my_total, sxx_total, syy_total, sxy_total]))
    tmmsss_array = np.array(tmmsss_array_list)
    tmmsss_array = tmmsss_array.reshape(tmmsss0_shape)
    return tmmsss_array


# ????
def mre(Ob, Fo):
    '''
    mre  精细化网格预报竞赛检验办法中的降水量定量相对误差检验指标
    :param Ob: 实况数据 不定长维度的numpy
    :param Fo: 测试数据 不定长维度的numpy
    :return: mre
    '''
    mre_list = []
    Fo_shape = Fo.shape
    Ob_shape = Ob.shape

    Ob_shpe_list = list(Ob_shape)
    size = len(Ob_shpe_list)
    ind = -size
    Fo_Ob_index = list(Fo_shape[ind:])
    if Fo_Ob_index != Ob_shpe_list:
        print('预报数据和观测数据维度不匹配')
        return

    s = Ob + Fo
    d = Ob - Fo
    if len(Fo_shape) == len(Ob_shape):
        if np.sum(s) == 0:
            return 0
        else:
            s1 = s[s > 0]
            d1 = d[s > 0]
            are0 = np.mean(np.abs(d1 / s1))
            return are0
    else:
        Ob_shpe_list.insert(0, -1)
        new_Fo_shape = tuple(Ob_shpe_list)
        new_s = s.reshape(new_Fo_shape)
        new_d = d.reshape(new_Fo_shape)
        new_Fo_shape = new_s.shape
        for line in range(new_Fo_shape[0]):
            s_piece = new_s[line, :]
            d_piece = new_d[line, :]
            if np.sum(s_piece) == 0:
                are0 = 0
            else:
                s1 = s_piece[s_piece > 0]
                d1 = d_piece[s_piece > 0]
                are0 = np.mean(np.abs(d1 / s1))
            mre_list.append(are0)
        mre_array = np.array(mre_list)
        shape = list(Fo_shape[:ind])
        mre_array = mre_array.reshape(shape)
        return mre_array


def mre_toar(toar_array):
    '''
    mre  精细化网格预报竞赛检验办法中的降水量定量相对误差检验指标
    :param toar_array: 包含命中空报和漏报的多维数组，其中最后一维长度为2，分别记录了（预报和观测值之和大于0样本数,各点相对误差绝对值总和）
    （预报和观测值之和大于0样本数、各点相对误差绝对值总和），它由toar返回
    :return:
    '''
    count = toar_array[..., 0] + 0
    if count.size == 1:
        if count == 0:
            mre0 = IV
        else:
            mre0 = toar_array[..., 1] / count
    else:
        count[count < 0] = 1e-10
        ar = toar_array[..., 1]
        mre0 = ar / count
        mre0[count < 1] = IV
    return mre0


def toar(Ob, Fo):
    '''
    相对误差检验指标的中间结果量
    :param Ob: 实况数据  任意维numpy数组
    :param Fo: 预测数据 任意维numpy数组,Fo.shape 和Ob.shape一致
    :return: 一维numpy数组，其内容依次为预报和观测值之和大于0样本数、各点相对误差绝对值总和
    '''

    toar_array_list = []
    Fo_shape = Fo.shape
    Ob_shape = Ob.shape

    Ob_shpe_list = list(Ob_shape)
    size = len(Ob_shpe_list)
    ind = -size
    Fo_Ob_index = list(Fo_shape[ind:])
    if Fo_Ob_index != Ob_shpe_list:
        print('预报数据和观测数据维度不匹配')
        return
    s = Ob + Fo
    d = Ob - Fo
    Ob_shpe_list.insert(0, -1)
    new_Fo_shape = tuple(Ob_shpe_list)
    new_Fo = Fo.reshape(new_Fo_shape)
    s = s.reshape(new_Fo_shape)
    d = d.reshape(new_Fo_shape)
    new_Fo_shape = new_Fo.shape
    for line in range(new_Fo_shape[0]):
        s_piece = s[line, :]
        d_piece = d[line, :]
        s1 = s_piece[s_piece > 0]
        d1 = d_piece[s_piece > 0]
        ar = np.sum(np.abs(d1 / s1))
        toar_array_list.append(np.array([s1.size, ar]))
    toar_array = np.array(toar_array_list)
    shape = list(Fo_shape[:ind])
    shape.append(2)
    toar_array = toar_array.reshape(shape)
    return toar_array


def nse(Ob, Fo):
    '''
    nse纳什系数, 常用于计算两个非正态序列的相对误差情况，
    :param Ob:实况数据 不定长维度的numpy
    :param Fo:测试数据 不定长维度的numpy
    :return:负无穷至1，最优值为1
    '''
    nse_array_list = []
    Fo_shape = Fo.shape
    Ob_shape = Ob.shape

    Ob_shpe_list = list(Ob_shape)
    size = len(Ob_shpe_list)
    ind = -size
    Fo_Ob_index = list(Fo_shape[ind:])
    if Fo_Ob_index != Ob_shpe_list:
        print('预报数据和观测数据维度不匹配')
        return

    mob = np.mean(Ob)
    qdob = np.mean(np.power(Ob - mob, 2))
    Ob_shpe_list.insert(0, -1)
    new_Fo_shape = tuple(Ob_shpe_list)
    new_Fo = Fo.reshape(new_Fo_shape)
    new_Fo_shape = new_Fo.shape
    for line in range(new_Fo_shape[0]):
        if qdob == 0:
            nse = IV
        else:
            nse = 1 - np.mean(np.power(Ob - new_Fo[line, :], 2)) / qdob
        nse_array_list.append(nse)
    if len(nse_array_list) == 1:
        return nse_array_list[0]
    else:
        nse_array = np.array(nse_array_list)
        shape = list(Fo_shape[:ind])
        nse_array = nse_array.reshape(shape)
        return nse_array


def nse_tase_tmmsss(tase_array, tmmsss_array):
    '''
    :param tase_array:
    :param tmmsss_array:
    :return:
    '''
    sxx = tmmsss_array[..., 3] + 0
    if sxx.size == 1:
        if sxx == 0:
            nse0 = IV
        else:
            nse0 = 1 - tase_array[..., 3] / tase_array[..., 0] / sxx
    else:
        sum = sxx + 0
        sum[sxx == 0] = 1e-10
        mse0 = tase_array[..., 3] / tase_array[..., 0]
        nse0 = 1 - mse0 / sum
        nse0[sxx == 0] = IV
    return nse0

def nse_tmmsss(tmmsss_array):
    '''
    :param tase_array:
    :param tmmsss_array:
    :return:
    '''

    sxx = tmmsss_array[..., 3] + 0
    syy = tmmsss_array[..., 4] + 0
    sxy = tmmsss_array[..., 5] + 0
    mx = tmmsss_array[..., 1] + 0
    my = tmmsss_array[..., 2] + 0
    exy = sxx + syy - 2 * sxy + (mx - my) * (mx - my)



    if sxx.size == 1:
        if sxx == 0:
            nse0 = IV
        else:
            nse0 = 1 - exy / sxx
    else:
        sum = sxx + 0
        sum[sxx == 0] = 1e-10
        mse0 = exy
        nse0 = 1 - mse0 / sum
        nse0[sxx == 0] = IV
    return nse0



def pmse(ob,fo):
    '''

    :param ob: 一维numpy数组
    :param fo:  一维或者两维numpy数组，当fo是两维时，表示有多种预报进行对比
    :return:  实数
    '''
    #如果fo只包含一个预报成员
    if len(ob.shape) == len(fo.shape):
        fo = fo[np.newaxis,:]
    nfo = fo.shape[0]
    list1 = []
    result_list = []
    nsta = ob.size
    #print("---")
    #print(nsta)
    #对多个预报进行循环
    for i in range(nfo):
        foi = fo[i,:]
        #ob3 = np.power(ob,0.33333)
        #fo3 = np.power(foi,0.333333)
        e = ob - foi
        e1 = e[e>=0]
        mse1 = np.sum(0.91*0.91* np.power(e1,0.66666))
        e2 = e[e<0]
        mse2 = np.sum(np.power(-0.7+0.57*np.power(-e2,0.3333),2))
        mse = mse1 + mse2
        result_list.append(mse/nsta)
    #result_mid = np.array(list1)
    result = np.array(result_list)

    return result



def pas(ob0,fo,grade_list = [0.1]):
    '''

    :param ob: 一维numpy数组
    :param fo:  一维或者两维numpy数组，当fo是两维时，表示有多种预报进行对比
    :return:  实数,pas评分
    '''
    #如果fo只包含一个预报成员
    if len(ob0.shape) == len(fo.shape):
        fo = fo[np.newaxis,:]
    nfo = fo.shape[0]
    list1 = []
    result_list = []
    ob_0 = ob0.flatten()
    #对多个预报进行循环
    for i in range(nfo):
        nsample = 0
        score = 0
        foi = fo[i,:].flatten()
        # 情况1， 实况降水u=0mm，预报降水x>0mm
        index0 = np.where((ob_0>=grade_list[0])|(foi>=grade_list[0]))
        #print(index0[0].size)

        ob  = ob_0[index0]
        foi = foi[index0]
        index1 = np.where((ob == 0)&(foi>0))
        score1 = 0
        if index1[0].size>0:
            fo1 = foi[index1]
            nsample += index1[0].size
            fo11 = fo1[fo1<5]         # 预报 0<x<5mm的情况
            score1  += 0.5* fo11.size
            fo12 = fo1[fo1>=5]        #预报 x>=5mm的情况
            if fo12.size>0:
                score1 += 0.5 * np.sum(np.exp(-(fo12 - 5) ** 2 / 25))
            # print("情况1：")
            # print(index1[0].size)
            #print(score1)

        # 情况2， 实况降水u>0mm，预报降水x=0mm
        index2 = np.where((ob > 0)&(ob<10)&(foi == 0))
        score2 = 0
        if index2[0].size>0:
            ob2 = ob[index2]
            nsample += index2[0].size
            score2 += 0.5 * np.sum(np.sin(0.05 * np.pi * (10 + ob2)))
            # print("情况2：")
            # print(index2[0].size)
            #print(score2)

        # 情况3， 实况降水0<u<5mm，预报降水x>0mm
        index3 = np.where((ob>0)&(ob<5)&(foi>0))
        score3 =0
        if index3[0].size >0:
            nsample += index3[0].size
            ob3 = ob[index3]
            fo3 = foi[index3]
            ob31 = ob3[fo3<ob3]   #情况3.1  0<x<u,0<u<5
            if ob31.size>0:
                fo31 = fo3[fo3 < ob3]
                score3 += np.sum(np.sin(0.05*np.pi*(10+ob31-fo31)))
            ob32 = ob3[np.where((fo3<5)&(ob3<=fo3))]   #情况3.2  0<x<5,0<u<x
            score3 += ob32.size
            ob33 = ob3[fo3>=5]  #情况3.3  x>=5,0<u<5
            if ob33.size>0:
                fo33 = fo3[fo3>=5]
                score3 += np.sum(np.exp(-(fo33-5)**2/25))
            # print("情况3：")
            # print(index3[0].size)
            #print(score3)

        # 情况4， 实况降水5<u<10mm，预报降水x>0mm
        index4 = np.where((ob>=5)&(ob<10)&(foi>0))
        score4 = 0
        if index4[0].size>0:
            nsample+= index4[0].size
            ob4 = ob[index4]
            fo4 = foi[index4]
            ob41 = ob4[fo4<ob4]  #情况4.1 0<x<u,5<u<10
            if ob41.size>0:
                fo41 = fo4[fo4<ob4]
                score4 += np.sum(np.sin(0.05*np.pi*(10+ob41-fo41)))
            ob42 = ob4[fo4>=ob4]  #情况4.2 u<x,5<u<10
            if ob42.size>0:
                fo42 = fo4[fo4>=ob4]
                score4 += np.sum(np.exp(-(fo42-ob42)**2/ob42**2))
            # print("情况4：")
            # print(index4[0].size)
            #print(score4)


        # 情况5， 实况降水u>=10mm，预报降水x>0mm
        index5 = np.where((ob >=10)  & (foi >= 0))
        score5 = 0
        if index5[0].size >0:
            nsample += index5[0].size
            ob5 = ob[index5]
            fo5 = foi[index5]
            ob51 = ob5[fo5<ob5]  #情况5.1, 0<=x<u,u>=10
            if ob51.size>0:
                fo51 = fo5[fo5<ob5]
                score5 += np.sum(np.sin(0.5*np.pi*fo51/ob51))
            ob52 = ob5[fo5>=ob5]  #情况5.2, u<=x,u>=10
            if ob52.size>0:
                fo52 = fo5[fo5>=ob5]
                score5 += np.sum(np.exp(-(fo52-ob52)**2/ob52**2))
            # print("情况5：")
            # print(index5[0].size)
            #print(score5)
        score = score1 + score2+score3+score4+score5
        #print(score)
        #计算
        if nsample==0:
            result1 = IV
        else:
            result1 = score/nsample
        list1.append([nsample,score])
        result_list.append(result1)

    #result_mid = np.array(list1)
    result = np.array(result_list)


    return result


if __name__=="__main__":
    import pandas as pd
    import datetime
    import meteva
    import xarray as xr

    rain01_ob = np.random.randn(1000)
    rain01_ob[rain01_ob < 0] = 0
    rain01_fo = np.random.randn(2, 1000)
    rain01_fo[rain01_fo < 0] = 0
    print("观测和预报的平均降水强度分别是：")
    result = ob_fo_precipitation_strength(rain01_ob, rain01_fo)  # 多个预报时 返回数组的size =  1+预报成员
    print(result)



#!/usr/bin/python3.6
# -*- coding:UTF-8 -*-
import datetime as datetime
import os as os
import numpy as np
from ..io import DataBlock_pb2
from ..io.GDS_data_service import GDSDataService
import meteva
import re
import pathlib

#获取最新的路径


#获取最新的gds路径
def get_latest_gds_path(dir,time,dtime,dt_cell = "hour",dt_step = 1,farthest = 240):
    dir1,filemodel = os.path.split(dir)
    file_list = get_gds_file_list_in_one_dir(dir1)
    for ddt in range(0,farthest,dt_step):
        if dt_cell.lower() == "hour":
            time1 = time - datetime.timedelta(hours=ddt)
        elif dt_cell.lower() == "minute":
            time1 = time - datetime.timedelta(minutes=ddt)
        filename = get_path(filemodel, time1, dtime + ddt, dt_cell)
        if filename in file_list:
            path = dir1 + "/" + filename
            return path

#获取最新的路径
def get_latest_path(dir,time,dt,dt_cell = "hour",dt_step = 1, farthest = 240):
    for ddt in range(0,240,dt_step):
        if dt_cell.lower()=="hour":
            time1 = time - datetime.timedelta(hours=ddt)
        elif dt_cell.lower()=="minute":
            time1 = time - datetime.timedelta(minutes=ddt)

        path = get_path(dir,time1,dt + ddt,dt_cell)
        if(os.path.exists(path)):
            return path
    return None


#获取路径
def get_path(dir,time,dt = None,dt_cell = "hour"):
    '''
    根据路径通配形式返回路径
    :param dir:
    :param time:
    :param dt:
    :param dt_cell:
    :return:
    '''
    if(dir.find("*")<0):
        return get_path_without_star(dir,time,dt,dt_cell)
    else:
        dir_0,filename_0 = os.path.split(dir)
        dir_1 = get_path_without_star(dir_0,time,dt,dt_cell)
        filename_1 = get_path_without_star(filename_0,time,dt,dt_cell)
        patten = dir_1  +"\\"+ filename_1
        patten = patten.replace("\\","/")
        patten = patten.replace("*","\\S+")
        if os.path.exists(dir_1):
            files = os.listdir(dir_1)
            for file in files:
                path = dir_1 + "/" + file
                path = path.replace("\\","/")
                match = re.match(patten,path)
                if match is not None:
                    return path
    return None



def get_path_without_star(dir,time,dt = None,dt_cell = "hour"):
    '''
    :param dir:
    :param time:
    :param dt:
    :param dt_cell:
    :return:
    '''
    if(dt is not None):
        if not (isinstance(dt,np.int16) or isinstance(dt,np.int32) or isinstance(dt,np.int64)or type(dt) == type(1)):
            if(dt_cell.lower()=="hour"):
                dt = int(dt.total_seconds() / 3600)
            elif(dt_cell.lower()=="minute"):
                dt = int(dt.total_seconds() / 60)
            elif(dt_cell.lower()=="day"):
                dt = int(dt.total_seconds() / (24*3600))
            else:
                dt = int(dt.total_seconds())
    else:
        dt = 0
    cdt2 = '%02d' % dt
    cdt3 = '%03d' % dt
    cdt4 = '%04d' % dt
    dir1 = dir.replace("TTTT",cdt4).replace("TTT",cdt3).replace("TT",cdt2)
    y4 = time.strftime("%Y")
    y2 = y4[2:]
    mo = time.strftime("%m")
    dd = time.strftime("%d")
    hh = time.strftime("%H")
    mi = time.strftime("%M")
    ss = time.strftime("%S")
    dir1 = dir1.replace("YYYY",y4).replace("YY",y2).replace("MM",mo).replace("DD",dd).replace("HH",hh).replace("FF",mi).replace("SS",ss)
    dir1 = dir1.replace(">","")
    return dir1

#创建路径
def creat_path(path):
    if path is not None:
        [dir,filename] = os.path.split(path)
        #if(not os.path.exists(dir)):
        #    os.makedirs(dir)
        pathlib.Path(dir).mkdir(parents=True,exist_ok=True)


#字符转换为datetime
def str_to_time(str0):
    return datetime.datetime.strptime(str0, '%Y%m%d%H%M')

#获取预报时效路径
def get_forecat_hour_of_path(path_model,path):
    ttt_index = path_model.find("TTT")
    if (ttt_index >= 0):
        ttt = int(path[ttt_index:ttt_index + 3])
    else:
        ttt = 0
    return ttt

#获取时间命名的路径
def get_time_of_path(path_model,path):

    yy_index = path_model.find("YYYY")
    if  yy_index < 0:
        yy_index = path_model.find("YY")
        if(yy_index <0):
            yy = 2000
        else:
            yy = int(path[yy_index: yy_index + 2])
    else:
        yy = int(path[yy_index: yy_index+4])

    mm_index = path_model.find("MM")
    if(mm_index >=0):
        mm = int(path[mm_index:mm_index+2])
    else:
        mm = 1

    dd_index = path_model.find("DD")
    if(dd_index>=0):
        dd = int(path[dd_index:dd_index + 2])
    else:
        dd = 1
    hh_index = path_model.find("HH")
    if(hh_index>=0):
        hh = int(path[hh_index:hh_index + 2])
    else:
        hh = 0
    ff_index = path_model.find("FF")
    if(ff_index>=0):
        ff = int(path[ff_index:ff_index + 2])
    else:
        ff = 0
    ss_index = path_model.find("SS")
    if(ss_index>=0):
        ss = int(path[ss_index:ss_index + 2])
    else:
        ss = 0
    return datetime.datetime(yy,mm,dd,hh,ff,ss)

#获取按照时间命名的文件夹名称
def get_dir_of_time(path_model,time):
    dir = os.path.split(get_path(path_model,time))[0]
    return dir

#获取按照时间命名的文件夹名称列表
def get_path_list_of_time(path_model,time):
    dir = get_dir_of_time(path_model,time) +"/"
    path_list = []
    if(os.path.exists(dir)):
        path_list = os.listdir(dir)
        for i in range(len(path_list)):
            path_list[i] = dir  + path_list[i]
    return path_list

#按照最新的时间命名文件路径
def get_time_nearest_path(path_model,time,max_seconds,path_list = None):
    if(path_list is None):
        path_list = get_path_list_of_time(path_model,time)
    dt_min = max_seconds
    nearest_path = None
    for path in path_list:
        time1 = get_time_of_path(path_model,path)
        dt = abs((time1 - time).total_seconds())
        if(dt < dt_min):
            dt_min = dt
            nearest_path = path
    return nearest_path

#按照上一个时效命名文件路径
def get_time_before_nearest_path(path_model,time,max_seconds,path_list = None):
    if(path_list is None):
        path_list = get_path_list_of_time(path_model,time)
    dt_min = max_seconds
    nearest_path = None
    for path in path_list:
        time1 = get_time_of_path(path_model,path)
        dt = (time - time1).total_seconds()
        if(dt < dt_min and dt >=0):
            dt_min = dt
            nearest_path = path
    return nearest_path

##按照下一个时效命名文件路径
def get_time_after_nearest_path(path_model,time,max_seconds,path_list = None):
    if(path_list is None):
        path_list = get_path_list_of_time(path_model,time)
    dt_min = max_seconds
    nearest_path = None
    for path in path_list:
        time1 = get_time_of_path(path_model,path)
        dt = (time1 - time).total_seconds()
        if(dt < dt_min  and dt >=0):
            dt_min = dt
            nearest_path = path
    return nearest_path

#以列表的形式返回根目录下所有文件的路径
def get_path_list_in_dir(root_dir,all_path = None):
    if not os.path.exists(root_dir):
        return []
    files = os.listdir(root_dir)
    if all_path is None:
        all_path = []
    for file in files:
        fi_d = os.path.join(root_dir,file)
        if os.path.isdir(fi_d):
            get_path_list_in_dir(fi_d,all_path)
        else:
            all_path.append(fi_d)

    return all_path

#返回根目录下nc格点文件的长名字
def get_path_of_grd_nc_longname(root_dir,time,dhour,nc_Fname,fhour_add):
    ruc_file = "{4}_IT_{0}_VT_{1}_FH_{2:0>3d}_AT_{3:0>3d}.nc".format(time.strftime("%Y%m%d%H"),
                                                                     (time + datetime.timedelta(hours=int(dhour))).strftime(
                                                                         "%Y%m%d%H"), dhour, fhour_add, nc_Fname)
    file = r"{0}\{1}\{2}".format(root_dir, time.strftime("%Y%m%d"), ruc_file)
    return file


def get_gds_file_list_in_one_dir(dir):
    dir = dir.replace("mdfs:///", "")

    ip,port = meteva.base.gds_ip_port
    service = GDSDataService(ip, port)
    # 获得指定目录下的所有文件
    status, response = service.getFileList(dir)
    MappingResult = DataBlock_pb2.MapResult()
    file_list = []
    if status == 200:
        if MappingResult is not None:
            # Protobuf的解析
            MappingResult.ParseFromString(response)
            results = MappingResult.resultMap
            # 遍历指定目录
            for name_size_pair in results.items():
                if (name_size_pair[1] != 'D'):
                    file_list.append(name_size_pair[0])
    return file_list

def get_gds_path_list_in_one_dir(dir):
    file_list = get_gds_file_list_in_one_dir(dir)
    path_list = []
    for file in file_list:
        path_list.append(dir + "/"+ file)
    return path_list

def exist_in_gds(path):
    dir,filename = os.path.split(path)
    file_list = get_gds_file_list_in_one_dir(dir)
    if filename in file_list:
        return True
    else:
        return False

def get_gds_all_dir(path,all_path,service = None):
    # 初始化GDS客户端

    if service is None:
        ip, port = meteva.base.gds_ip_port
        service = GDSDataService(ip, port)
    # 获得指定目录下的所有文件
    path = path.replace("mdfs:///", "")
    status, response = service.getFileList(path)
    MappingResult = DataBlock_pb2.MapResult()
    if status == 200:
        if MappingResult is not None:
            # Protobuf的解析
            MappingResult.ParseFromString(response)
            results = MappingResult.resultMap
            # 遍历指定目录
            contain_dir = False
            for name_size_pair in results.items():
                if (name_size_pair[1] == 'D'):
                    contain_dir = True
                    path1 = '%s%s%s' % (path, "/" , name_size_pair[0])
                    if(path1[0:1] == "/"):
                        path1 = path1[1:]
                    if(path1[0:1] == "/"):
                        path1 = path1[1:]
                    get_gds_all_dir(path1,all_path,service)
                    #print(name_size_pair[0])
            if(not contain_dir):
                all_path.append(path)



def get_dati_of_path(path):
    try:
        dir,filename = os.path.split(path)
        filename0 = os.path.splitext(filename)[0]
        a = int(filename0[0:2])
        b = int(filename0[2:4])
        if a ==20:
            if b >12:
                pass
            else:
                filename0 = "20" + filename0
        elif a == 19:
            if b >12:
                pass
            else:
                filename0 = "20" + filename0
        else:
            filename0 = "20" + filename0

        dati = str_to_time(filename0)
        return dati
    except:
        return None

#以列表的形式返回根目录下所有文件的路径
def get_during_path_list_in_dir(root_dir,all_path = None,start = None,end = None):
    start_time = None
    if start is None:
        start_time = datetime.datetime(1900,1,1,0,0)
    else:
        start_time = start
    if end is None:
        end_time = datetime.datetime.now()
    else:
        end_time = end
    time_compair = False
    if start is not None or end is not None:
        time_compair = True

    if not os.path.exists(root_dir):
        return []
    files = os.listdir(root_dir)
    if all_path is None:
        all_path = []
    for file in files:
        fi_d = os.path.join(root_dir,file)
        if os.path.isdir(fi_d):
            if time_compair:
                dati = get_dati_of_path(fi_d)
                if dati is None:
                    get_during_path_list_in_dir(fi_d, all_path, start, end)
                else:
                    if dati >= start_time and dati <= end_time:
                        get_during_path_list_in_dir(fi_d,all_path,start,end)
            else:
                get_during_path_list_in_dir(fi_d, all_path, start, end)
        else:
            all_path.append(fi_d)
    return all_path


import math


def lengths_psp(x, squared=False):
    lengths2 = (x['x1'] - x['x0']) ** 2 + (x['y1'] - x['y0']) ** 2
    if squared:
        return lengths2
    else:
        return math.sqrt(lengths2)



# -*- coding: utf-8 -*-
"""
Created on Fri Jan 22 18:12:20 2021
"""
import numpy as np
import math
import matplotlib.pyplot as plt
import meteva

def rdist(Id, loc):
    # 计算两点间欧式距离
    a = Id[:,0]
    b = Id[:,1]
    lat_lon0 = loc[a]
    lat_lon1 = loc[b]
    sr=np.cos(0.5*(lat_lon0[:,0]+lat_lon1[:,0])*math.pi/180)
    d1=(lat_lon0[:,1] - lat_lon1[:,1])*sr
    d2=(lat_lon0[:,0] - lat_lon1[:,0])
    dd = np.sqrt(d1*d1+d2*d2)
    dd *= meteva.base.dis_1_degree
    return dd


def dv(sta,dmax = None, q = 1):
    '''

    :param sta: 输入单个层次、单个时刻、单个时效的站点数据
    :param q:  q=1对应误差绝对值，q=2对应误差的平方
    :param dmax:  距离大于dmax的样本将被滤除
    :return:  N×（站点数据列数+1）  的二维数组，第1列为距离值，之后为sta中每个数据列中样本的
    '''
    sta1 = meteva.base.not_IV(sta)
    loc =sta1[["lat","lon"]].values
    n = len(loc)
    names = meteva.base.get_stadata_names(sta)
    aa = np.tile(np.arange(0, n), n)
    bb = np.repeat(np.arange(0, n), n)
    ind = aa > bb
    Id = np.vstack((aa, bb)).T[ind, :]
    d = rdist(Id,loc)
    od = np.argsort(d)
    d = d[od]
    index = None
    if dmax is not None:
        index = np.where(d<dmax)
        d = d[index]
    dv_list = [d]

    for name in names:
        y = sta1[name].values
        if q == 1:
            v = np.abs(y[Id[:, 0]] - y[Id[:, 1]])
        else:
            v = (y[Id[:, 0]] - y[Id[:, 1]]) ** q
        v = v[od]
        if index is not None:
            v = v[index]
        dv_list.append(v)
    dv_array = np.vstack(tuple(dv_list)).T
    return dv_array

def dv_merge(dv_array1,dv_array2):
    '''
    将两个变率样本结果合并
    :param dv_array1: 变率样本矩阵2
    :param dv_array2: 变率样本矩阵2
    :return:
    '''
    dv_array = np.vstack((dv_array1,dv_array2))
    return dv_array

def plot_dv(dv_array,dmax = None,grade_count = 10
            , member_list=None, width=None, height=None, xlabel="距离（km）",ylabel1="变化幅度(统计值）",ylabel2="变化幅度(箱须图）"
            ,title = "Variogram(变差图)",
            dpi=300, show=False, save_path=None, sup_fontsize=10
            ):

    '''
    将变率样本矩阵绘制成图形产品
    :param dv_array: 变率样本矩阵
    :param grade_count: 绘图时，将样本按距离分为 grade_count组进行统计
    :param dmax:         统计的最大距离值，超出的部分会被剔除
    :param member_list:    legend 的名称内容
    :param width:         画面的宽度
    :param height:        画面的高度
    :param xlabel:
    :param ylabel1:
    :param ylabel2:
    :param dpi:
    :param show:
    :param save_path:
    :param sup_fontsize:
    :return:
    '''
    if dmax is None:
        dmax = np.max(dv_array[:,0])
    grade = dmax / grade_count
    grade_list = (grade * (np.arange(grade_count)+1)).astype(np.int16)
    if height is None:
        height = 4
    if width is None:
        width = meteva.base.tool.plot_tools.caculate_axis_width(grade_list,sup_fontsize*1.5,legend_num=1)
    fig = plt.figure(figsize=(width, height),dpi = dpi)

    grid_plt = plt.GridSpec(9, 1, hspace=0)
    cmv_array, dmax = dv_statistic(dv_array, dmax, grade_count)

    ax0 = plt.subplot(grid_plt[0:4, 0])
    plt.title(title,fontsize = sup_fontsize)
    v_count = dv_array.shape[-1] - 1
    label = []
    if member_list is None:

        label = ["观测"]
        for i in range(1, v_count):
            label.append('预报' + str(i))
    else:
        label.extend(member_list)

    color_list = meteva.base.color_tools.get_color_list(v_count)
    x = np.arange(0.5, grade_count)

    plt.yticks(fontsize=sup_fontsize * 0.8)
    plt.ylabel(ylabel1, fontsize=sup_fontsize * 0.9)
    max_list = []
    for v in range(v_count):
        plt.plot(x, cmv_array[1, :, v], c=color_list[v], label=label[v])
        var = np.sqrt(cmv_array[2, :, v])
        plt.plot(x, cmv_array[1, :, v] - var, "--", c=color_list[v], linewidth=0.7)
        max_list.append(np.max(cmv_array[1, :, v] + var))
        plt.plot(x, cmv_array[1, :, v] + var, "--", c=color_list[v], linewidth=0.7)
    maxv = np.max(np.array(max_list))
    plt.ylim(0, maxv * 1.2)
    plt.legend(loc=2, ncol=4, fontsize=sup_fontsize * 0.8)

    d = dv_array[:, 0]
    boxgroup = []
    combineData = []

    tcount = []


    for i in range(grade_count):
        index = np.where((d> i*grade) & (d<=(i+1)* grade))
        count = len(index[0])
        tcount.append(count)
        for v in range(1,v_count+1):
            dat = dv_array[index,v].squeeze()
            combineData.append(dat)
            bg = np.ones(count) * i + 0.5
            boxgroup.append(bg)



    ax1 = plt.subplot(grid_plt[4:8, 0])

    red_square = dict(markerfacecolor='k',markeredgecolor = "k", markersize = 0.5)
    medianprops = dict(color = "k")
    bplot = ax1.boxplot(combineData,boxgroup, showfliers=True, patch_artist=True, sym='.',flierprops=red_square,medianprops = medianprops)
    color_list =meteva.base.color_tools.get_color_list(v_count)

    for i, item in enumerate(bplot["boxes"]):
        ci  = i%v_count
        item.set_facecolor(color_list[ci])
        item.set_edgecolor(color_list[ci])




    plt.subplots_adjust(left=0.5 / width, right=1 - 0.1 / width)
    plt.ylabel(ylabel2, fontsize=sup_fontsize * 0.9)
    plt.yticks(fontsize=sup_fontsize * 0.8)
    #plt.legend(fontsize=sup_fontsize * 0.8, ncol=3, loc="upper center")
    #plt.xlim(0,0.1)
    maxv = np.max(dv_array[:,1:])
    plt.ylim(0,maxv*1.2)
    #plt.legend(loc =2,ncol = 4,fontsize = sup_fontsize * 0.8)
    ax2 = plt.subplot(grid_plt[8, 0])
    plt.bar(x, tcount, width=0.5, label="样本数",color = "grey")
    plt.xlabel(xlabel, fontsize=sup_fontsize * 0.9)
    plt.xticks(x+0.5,grade_list)
    plt.xlim(0, len(tcount))
    plt.ylim(0, np.max(tcount) * 1.5)
    plt.ylabel("样本数",fontsize = sup_fontsize * 0.9)
    plt.grid()

    if save_path is None:
        show = True
    else:
        plt.savefig(save_path,bbox_inches='tight')
        print("检验结果已以图片形式保存至" + save_path)
    if show is True:
        plt.show()
    plt.close()



def dv_statistic(dv_array,dmax = None,grade_count=10):
    '''
    从变率样本矩阵中，统计出每一段距离范围内，平均的变率，以及变率值本身的方差
    :param dv_array:
    :param dmax:
    :param grade_count:
    :return:
    '''
    if dmax is None:
        dmax = np.max(dv_array[:,0])
    d_step = dmax/grade_count
    v_count = dv_array.shape[1] - 1

    #c：count, m ：mean， v :var
    cmv_array = np.zeros((3,grade_count,v_count))
    d = dv_array[:, 0]
    for i in range(grade_count):
        index = np.where((d> i*d_step) & (d<=(i+1)* d_step))
        count = len(index[0])
        if count>0:
            for j in range(v_count):
                cmv_array[0,i,j] = count
                v = dv_array[index,j+1]
                cmv_array[1,i,j] = np.mean(v)
                cmv_array[2,i, j] = np.var(v)

    #返回dmax，是方便后续程序确切了解组成采用dmax参数值
    return cmv_array,dmax

def merge_dv_statistic(cmv_array1,cmv_array2):
    '''
    将统计结果合并
    :param cmv_array1:
    :param cmv_array2:
    :return:
    '''
    count_array,mean_array,var_array = meteva.base.math_tools.ss_iteration(cmv_array1[0,...],cmv_array1[1,...],cmv_array1[2,...],
                                                                           cmv_array2[0,...],cmv_array2[1,...],cmv_array2[2,...])

    cmv_array = np.vstack((count_array,mean_array,var_array))
    return cmv_array


def plot_dv_statistic(cmv_array,dmax,member_list = None,width = None,height = None,xlabel = "距离（km）",ylabel = "变化幅度",
                      title = "Variogram(变差图)",
                      dpi = 300,show =False,save_path = None,sup_fontsize = 12):
    '''
    绘制变率统计结果
    :param cmv_array:
    :param dmax:
    :param member_list:
    :param width:
    :param height:
    :param xlabel:
    :param ylabel:
    :param dpi:
    :param show:
    :param save_path:
    :param sup_fontsize:
    :return:
    '''
    if height is None:
        height = 3

    grade_count = cmv_array.shape[-2]
    grade = dmax / grade_count

    grade_list = (grade * (np.arange(grade_count)+1)).astype(np.int16)
    grade_list_mid = (grade * (np.arange(grade_count)+0.5)).astype(np.int16)
    if width is None:
        width = meteva.base.tool.plot_tools.caculate_axis_width(grade_list,sup_fontsize*1.5,legend_num=1)
    fig = plt.figure(figsize=(width, height),dpi = dpi)

    v_count = cmv_array.shape[-1]
    label = []
    if member_list is None:

        label = ["观测"]
        for i in range(1,v_count):
            label.append('预报' + str(i))
    else:
        label.extend(member_list)

    color_list =meteva.base.color_tools.get_color_list(v_count)

    grid_plt = plt.GridSpec(5, 1, hspace=0)
    ax1 = plt.subplot(grid_plt[0:4, 0])
    plt.title(title,fontsize = sup_fontsize)
    plt.yticks(fontsize = sup_fontsize * 0.8)
    plt.ylabel(ylabel,fontsize = sup_fontsize * 0.9)
    max_list =[]
    for v in range(v_count):
        plt.plot(grade_list_mid,cmv_array[1,:,v],c = color_list[v],label= label[v])
        var = np.sqrt(cmv_array[2,:,v])
        plt.plot(grade_list_mid,cmv_array[1,:,v] - var,"--",c = color_list[v],linewidth =0.5)
        max_list.append(np.max(cmv_array[1,:,v] + var))
        plt.plot(grade_list_mid,cmv_array[1,:,v] + var,"--",c = color_list[v],linewidth =0.5)


    plt.xlim(0,grade_list[-1])
    maxv = np.max(np.array(max_list))
    plt.ylim(0,maxv*1.2)
    plt.legend(loc =2,ncol = 4,fontsize = sup_fontsize * 0.8)
    plt.setp(ax1.get_xticklabels(), visible=False)
    ax2 = plt.subplot(grid_plt[4, 0], sharex=ax1)
    plt.yticks(fontsize=sup_fontsize * 0.8)
    plt.ylabel("样本数",fontsize = sup_fontsize * 0.9)
    plt.xlim(0,grade_list[-1])
    plt.xlabel(xlabel,fontsize = sup_fontsize * 0.9)
    plt.bar(grade_list_mid, cmv_array[0,:,0], width=grade * 0.5,color = "grey")
    #plt.xlabel(xlabel=grade_list)

    plt.xticks(grade_list,fontsize = sup_fontsize * 0.8)
    if save_path is None:
        show = True
    else:
        plt.savefig(save_path,bbox_inches='tight')
        print("检验结果已以图片形式保存至" + save_path)
    if show is True:
        plt.show()
    plt.close()


def vgm_sta(sta,dmax,grade_count = 10,q = 1,xlabel = "距离（km）",ylabel = "变化幅度",title = "Variogram(变差图)",
                  dpi = 300,show =False,save_path = None,sup_fontsize  = 10,width = None,height = None):
    '''
    统计函数的主入口，    统计并绘制要素值随距离变化的程度，只包括统计值曲线
    :param sta:
    :param q:
    :param dmax:
    :param grade_count:
    :param width:
    :param height:
    :param xlabel:
    :param ylabel:
    :param dpi:
    :param show:
    :param save_path:
    :param sup_fontsize:
    :return:
    '''
    data_names = meteva.base.get_stadata_names(sta)
    sta_list = meteva.base.split(sta)

    dv_array = dv(sta_list[0],dmax = dmax,q = q)
    cmv_array,dmax  = dv_statistic(dv_array,dmax = dmax,grade_count=grade_count)
    n_sta = len(sta_list)
    if n_sta >1:
        dv_array1 = dv(sta_list[0], dmax=dmax, q=q)
        cmv_array1, _ = dv_statistic(dv_array1, dmax=dmax, grade_count=grade_count)
        cmv_array= merge_dv_statistic(cmv_array,cmv_array1)

    plot_dv_statistic(cmv_array,dmax,member_list=data_names,width=width,height=height,
                      xlabel = xlabel,ylabel = ylabel,
                      dpi = dpi,show = show,title = title,save_path=save_path,sup_fontsize=sup_fontsize)
    return cmv_array


def vgm_sta_box(sta,dmax,grade_count = 10,q = 1,xlabel = "距离（km）",ylabel1="变化幅度(统计值）",ylabel2="变化幅度(箱须图）",title = "Variogram(变差图)",
                  dpi = 300,show =False,save_path = None,sup_fontsize  = 10,width = None,height = None):
    '''
    统计并绘制要素值随距离变化的程度，包括统计值和箱须图
    :param sta:
    :param dmax:
    :param q:
    :param grade_count:
    :param width:
    :param height:
    :param xlabel:
    :param ylabel1:
    :param ylabel2:
    :param dpi:
    :param show:
    :param save_path:
    :param sup_fontsize:
    :return:
    '''
    data_names = meteva.base.get_stadata_names(sta)
    sta_list = meteva.base.split(sta)
    dv_array = dv(sta_list[0],dmax = dmax,q = q)

    n_sta = len(sta_list)
    if n_sta >1:
        dv_array1 = dv(sta_list[0], dmax=dmax, q=q)
        dv_array = dv_merge(dv_array,dv_array1)
    plot_dv(dv_array,dmax,grade_count,member_list=data_names,width=width,height=height,
                      xlabel = xlabel,ylabel1 = ylabel1,ylabel2= ylabel2,title = title,
                      dpi = dpi,show = show,save_path=save_path,sup_fontsize=sup_fontsize)

    return dv_array
















import numpy as np
import scipy.stats as st
#from .regress2 import regress2



def sma(data):
    x = data['x']
    y = data['y']
    slope =0
    intercept = 0
    try:
        #coeff = regress2(x, y, _method_type_2="reduced major axis")
        slope, intercept, r_value, p_value, std_err = st.linregress(x, y)
    except:
        pass
    while np.isnan(slope) or slope==0:
        maxx = np.max(x) - np.min(x)
        maxy = np.max(y) - np.min(y)
        maxd = max(maxx,maxy)
        x1 = np.random.randn(len(x))* 0.01 * maxd + x
        y1 = np.random.randn(len(y))* 0.01 * maxd + y
        #coeff = regress2(x1, y1, _method_type_2="reduced major axis")
        slope, intercept, r_value, p_value, std_err = st.linregress(x1, y1)

    frome = (min(y + slope * x) - intercept) / (2 * slope)
    to = (max(y + slope * x) - intercept) / (2 * slope)
    l = {}
    l["slope"] = slope
    l["intercept"] = intercept
    #l['coef'] = coeff
    l['from'] = frome
    l['to'] = to
    return l


# def sma_bak(data, log="", method="SMA",
#         type="elevation", alpha=0.05, slope_test=None, elev_test=None,
#         multcomp=False, multcompmethod="default", robust=False, V=np.zeros([2, 2]), n_min=3, quiet=False, *params):
#     x = data['x']
#     y = data['y']
#     coeff = None
#     try:
#         coeff = regress2(x, y, _method_type_2="reduced major axis")
#     except:
#         pass
#     while coeff is None or (np.isnan(coeff["predict"][0])) or coeff['slope']==0:
#         maxx = np.max(x) - np.min(x)
#         maxy = np.max(y) - np.min(y)
#         maxd = max(maxx,maxy)
#         x1 = np.random.randn(len(x))* 0.01 * maxd + x
#         y1 = np.random.randn(len(y))* 0.01 * maxd + y
#         #coeff = regress2(x1, y1, _method_type_2="reduced major axis")
#         coeff = regress2(x1, y1, _method_type_2="major axis")
#         #print(y)
#     #print(coeff)
#     slope = coeff['slope']
#     intercept = coeff['intercept']
#
#     frome = (min(y + slope * x) - intercept) / (2 * slope)
#     to = (max(y + slope * x) - intercept) / (2 * slope)
#     l = {}
#     l['coef'] = coeff
#     l['from'] = frome
#     l['to'] = to
#     return l



import meteva
import pandas as pd
import numpy as np

def p2p_vto01(sta_all,threshold = 1e-30,compair = ">="):
    '''

    :param sta_fo_all: 将预报数据处理成0-1数据
    :param threshold: 大于等于该阈值参数的站点值将转换成1，否则转换成0，缺省值仍然保持为缺省值
    :return:
    '''
    if compair not in [">=",">","<","<="]:
        print("compair 参数只能是 >=   >  <  <=  中的一种")
        return

    sta_happen_all = sta_all.copy()
    fo_names = meteva.base.get_stadata_names(sta_all)

    if isinstance(threshold,pd.DataFrame):
        sta_compair = meteva.base.combine_expand_IV(sta_happen_all,threshold)
        print(sta_compair.iloc[-10:-1,:])
        for i in range(len(fo_names)):
            name = fo_names[i]
            hap = np.zeros(len(sta_all.index))
            if compair == ">=":
                hap[sta_compair.loc[:,name].values >= sta_compair.iloc[:,-1].values] = 1
            elif compair =="<=":
                hap[sta_compair.loc[:,name].values <= sta_compair.iloc[:,-1].values] = 1
            elif compair ==">":
                hap[sta_compair.loc[:,name].values > sta_compair.iloc[:,-1].values] = 1
            else:
                hap[sta_compair.loc[:,name].values < sta_compair.iloc[:,-1].values] = 1

            sta_happen_all.loc[:, name] =hap[:]
            sta_happen_all.loc[sta_compair.loc[:,name].values == meteva.base.IV, name] = meteva.base.IV
    else:
        for i in range(len(fo_names)):
            name = fo_names[i]
            if compair == ">=":
                sta_happen_all.loc[:,name] = 0 + (sta_happen_all.loc[:,name] >= threshold)
            elif compair ==">":
                sta_happen_all.loc[:,name] = 0 + (sta_happen_all.loc[:,name] > threshold)
            elif compair =="<=":
                sta_happen_all.loc[:,name] = 0 + (sta_happen_all.loc[:,name] <= threshold)
            else:
                sta_happen_all.loc[:,name] = 0 + (sta_happen_all.loc[:,name] < threshold)
            sta_happen_all.loc[sta_all[name] == meteva.base.IV, name] = meteva.base.IV
    return sta_happen_all


def p2a_vto01(sta_all,r = 40,threshold = 1e-30,compair = ">="):
    '''

    :param sta_ob_all: 原始输入的观测站点数据,其中可以包含多个时刻的观测数据
    :param r: 临域半径，默认值为40km
    :param threshold: 半径璠r范围内观测值大于threshold的样本记为发生。
    :return: 采用临域法处理后的观测数据,其站点列表和sta_ob_all 一致
    '''

    # 将sta_ob_all 拆解成一个列表，列表中的每个元素为仅包含单个时刻观测的站点数据
    sta_all_01 = p2p_vto01(sta_all,threshold=threshold,compair = compair)
    sta_01_list = meteva.base.split(sta_all_01)
    npic = len(sta_01_list)
    sta_hapend_list = []
    for i in range(npic):
        sta_one = sta_01_list[i]                        # 提取单个时刻的观测数据
        sta_one_no_iv = meteva.base.not_IV(sta_one)     # 将其中的缺省数据删除
        #对于sta_one中的每个点，从不包含缺省值的数据sta_one_no_iv中找到半径r范围内的最大值
        #查询结果sta_happen的站点顺序和sta_one 一致
        sta_happen = meteva.base.max_in_r_of_sta(sta_one,r = r, sta_from = sta_one_no_iv)
        sta_hapend_list.append(sta_happen)
    sta_happen_all = pd.concat(sta_hapend_list,axis=0)
    return sta_happen_all



def sta_ob_fos_to_01(sta_of,sta_fo_list = None,r = 40, threshold_ob = 1e-30, threhold_fo = 1e-30):
    if sta_fo_list is None:
    # 将观测转为dtim0 = 0的形式
        data_names = meteva.base.get_stadata_names(sta_of)
        sta_ob = meteva.base.sele_by_para(sta_of, member=[data_names[0]])
        dhours = sta_ob["dtime"].values
        sta_ob = meteva.base.move_fo_time(sta_ob,dhours)
        sta_ob = sta_ob.drop_duplicates()
        sta_ob_01 = p2a_vto01(sta_ob,r= r,threshold = threshold_ob)
        sta_fo = meteva.base.sele_by_para(sta_of, member=[data_names[0]])
        sta_fo_01 = p2p_vto01(sta_fo,threshold= threhold_fo)
        sta_all_01 = meteva.base.combine_on_obTime_id(sta_ob_01,sta_fo_01)
    else:
        sta_ob_01 = p2a_vto01(sta_of, r=r, threshold=threshold_ob)
        sta_fo_01_list =[]
        for sta_fo in sta_fo_list:
            sta_fo_01 = p2p_vto01(sta_fo,threshold= threhold_fo)
            sta_fo_01_list.append(sta_fo_01)
        sta_all_01 = meteva.base.combine_on_obTime_id(sta_ob_01,sta_fo_01_list)
    return sta_all_01




from meteva.method.space.fuzzy_logic.lib.hoods2d import hoods2d
#import pyreadr
from meteva.method.space.fuzzy_logic.lib.make_spatialVx import make_spatialVx


def fuzzy_fss(input_object, levels, time_point=1,
          obs=0, model=0, max_n=None, rule=">=", verbose=False):
    result = hoods2d(input_object, which_methods=["fss"], time_point=time_point,
                     obs=obs, model=model, levels=levels, max_n=max_n, rule=rule, verbose=verbose)["fss"]
    result["thresholds"] = input_object["thresholds"][0]
    result["levels"] = levels
    return result


if __name__ == '__main__':
    import meteva.base as meb
    grid1 = meb.grid([100, 120, 0.05], [24, 40, 0.05])
    path_ob = r'H:\test_data\input\mem\mode\ob\rain03\20072611.000.nc'
    path_fo_03 = r'H:\test_data\input\mem\mode\ec\rain03\20072608.003.nc'
    path_fo_27 = r'H:\test_data\input\mem\mode\ec\rain03\20072508.027.nc'
    grd_ob = meb.read_griddata_from_nc(path_ob, grid=grid1, time="2020072611", dtime=0, data_name="OBS")
    grd_fo_03 = meb.read_griddata_from_nc(path_fo_03, grid=grid1, time="2020072608", dtime=3, data_name="ECMWF")
    grd_fo_27 = meb.read_griddata_from_nc(path_fo_27, grid=grid1, time="2020072508", dtime=27, data_name="ECMWF")

    obs_array = grd_ob.values.squeeze()
    fst_array = grd_fo_03.values.squeeze()
    hold = make_spatialVx(obs_array,fst_array)
    #look = joint(hold, levels=[1, 3, 9, 17, 33, 65, 129, 257], verbose=True)
    look = fuzzy_fss(hold, levels=[3], verbose=True)
    print(look)



import numpy as np
import meteva
import matplotlib.pyplot as plt
from meteva.method.yes_or_no.score import pofd_hfmc, pod_hfmc
from meteva.base.tool.plot_tools import set_plot_IV
from meteva.base import IV

import math


def reliability(Ob, Fo, grade_count=10, member_list=None, vmax = None,log_y = False,save_path=None,show = False,dpi = 300, title="可靠性图",
                sup_fontsize =10,width = None,height = None):
    '''
    :param Ob:
    :param Fo:
    :param save_path:
    :return:
    '''
    hnh_array = meteva.method.hnh(Ob, Fo, grade_count)

    result_dict = reliability_hnh(hnh_array, member_list=member_list,vmax = vmax,log_y = log_y,dpi = dpi, save_path=save_path,show = show, title=title,
                    width=width, height=height, sup_fontsize=sup_fontsize)
    return result_dict

def reliability_hnh(hnh_array,  member_list=None,vmax = None,log_y = False, save_path=None,show = False,dpi = 300, title="可靠性图",
                    sup_fontsize= 10,width = None,height = None):
    '''
    根据中间结果计算
    :param th:
    :param save_path:
    :return:
    '''

    grade_count = hnh_array.shape[-2]
    grade = 1 / grade_count
    if grade_count < 1:
        print('grade_count输入错误，不能小于1')
        return
    grade_list = np.arange(0, 1, grade).tolist()
    grade_list.append(1.1)
    hnh_array = hnh_array.reshape((-1, len(grade_list) - 1, 2))
    new_hnh_array_shape = hnh_array.shape
    label = []
    legend_num = new_hnh_array_shape[0]
    if member_list is None:
        if legend_num== 1:
            label.append('预报')
        else:
            for i in range(legend_num):
                label.append('预报' + str(i + 1))
    else:
        label.extend(member_list)



    color_list = meteva.base.tool.color_tools.get_color_list(legend_num)

    if width is None:
        width = 5.5
    if height is None:
        height = 6.1

    ngrade = len(hnh_array[0, :, 0])
    grade = 1 / ngrade
    x = np.arange(grade / 2, 1, grade)
    return_dict = {}
    return_dict["p_fo"] =x
    return_dict["p_ob"] = {}
    return_dict["sample_count"] = {}
    for line in range(new_hnh_array_shape[0]):
        total_grade_num = hnh_array[line, :, 0]
        observed_grade_num = hnh_array[line, :, 1]

        total_num = np.sum(total_grade_num)
        under = np.zeros_like(total_grade_num)
        under[:] = total_grade_num[:]
        under[total_grade_num == 0] = 1
        ob_rate = observed_grade_num / under
        ob_rate[total_grade_num == 0] = IV
        ob_rate_noIV = set_plot_IV(ob_rate)
        return_dict["p_ob"][line] =ob_rate
        ob_rate[total_grade_num == 0] = np.nan
        index_iv = np.where(total_grade_num == 0)
        line_x = np.arange(0, 1.00, grade)
        prefect_line_y = np.arange(0, 1.00, grade)
        climate_line_y = np.ones_like(line_x) * np.sum(observed_grade_num) / total_num


        if line == 0:
            fig = plt.figure(figsize=(width,height),dpi = dpi)
            grid_plt = plt.GridSpec(5, 1, hspace=0)
            ax1 = plt.subplot(grid_plt[0:4, 0])
            plt.plot(line_x, prefect_line_y, '--', label="完美", color="k")
            plt.plot(line_x, climate_line_y, ':', label="无技巧", color="k")
        plt.plot(x, ob_rate_noIV, "--", linewidth=0.5, color="k")
        x_iv = x[index_iv[0]]
        ob_rate_noIV_iv = ob_rate_noIV[index_iv[0]]
        plt.plot(x_iv, ob_rate_noIV_iv, "x", color='k')
        plt.plot(x, ob_rate, marker=".", markersize="10", label=label[line], color=color_list[line])

        plt.setp(ax1.get_xticklabels(), visible=False)
        plt.ylim(0.0, 1)
        plt.ylabel("观测频率",fontsize = sup_fontsize * 0.9)
        plt.legend(loc=2,fontsize = sup_fontsize * 0.9)
        plt.title(title,fontsize = sup_fontsize)
    bar_width = 0.8 / (grade_count * (legend_num+2))
    for line in range(new_hnh_array_shape[0]):
        total_grade_num = hnh_array[line, :, 0]
        observed_grade_num = hnh_array[line, :, 1]
        ngrade = len(total_grade_num)
        grade = 1 / ngrade
        under = np.zeros_like(total_grade_num)
        under[:] = total_grade_num[:]
        under[total_grade_num == 0] = 1
        ob_rate = observed_grade_num / under
        ob_rate[total_grade_num == 0] = IV
        ob_rate[total_grade_num == 0] = np.nan

        x = np.arange(grade / 2, 1, grade)
        if line == 0:
            ax2 = plt.subplot(grid_plt[4, 0], sharex=ax1)
        #x1 = x - 0.01 + (line + 1.5) * 0.01

        x1 = x + (line - legend_num/2 + 0.5) * bar_width
        return_dict["sample_count"][line] = total_grade_num
        plt.bar(x1, total_grade_num, width=bar_width*0.8,color=color_list[line])
        plt.ylabel("样本数",fontsize = sup_fontsize * 0.9)
        if log_y: plt.yscale('log')
        plt.xlim(0.0, 1.0)
        plt.xticks(np.arange(0.1, 1.01, 0.1))
        plt.xlabel("预报概率",fontsize = sup_fontsize * 0.9)
        if vmax is not None:
            plt.ylim(0,vmax)
    if save_path is None:
        show = True
    else:
        plt.savefig(save_path,bbox_inches='tight')
        print("检验结果已以图片形式保存至" + save_path)
    if show is True:
        plt.show()
    plt.close()
    return return_dict


def roc(Ob, Fo, grade_count=10, member_list=None, save_path=None,show = False,dpi = 300, title="ROC图",
        sup_fontsize =10,width = None,height = None):
    '''

    :param Ob:
    :param Fo:
    :param grade_count:
    :param save_path:
    :return:
    '''
    hnh_array = meteva.method.hnh(Ob, Fo, grade_count)
    return_dict = roc_hnh(hnh_array,  member_list=member_list, save_path=save_path,show = show,dpi = dpi, title=title,
            width=width, height=height, sup_fontsize=sup_fontsize)
    return return_dict

def roc_hfmc(hfmc_array, member_list=None, save_path=None,show = False,dpi = 300, title="ROC图",
            sup_fontsize =10,width = None,height = None):
    '''

    :param hfmc:
    :param save_path:
    :return:
    '''
    if width is None:
        width = 5
    if height is None:
        height = 5
    fig = plt.figure(figsize=(width, height),dpi = dpi)
    grade_count = hfmc_array.shape[-2]
    grade = 1 / grade_count
    if grade_count < 1:
        print('grade_count输入错误，不能小于1')
        return
    grade_list = np.arange(0, 1, grade).tolist()
    grade_list.append(1.1)
    shape = list(hfmc_array.shape)
    new_hfmc_array = hfmc_array.reshape((-1, len(grade_list) - 1, 4))
    new_hfmc_array_shape = new_hfmc_array.shape
    label = []
    if member_list is None:
        if new_hfmc_array_shape[0] == 1:
            label.append('预报')
        else:
            for i in range(new_hfmc_array_shape[0]):
                label.append('预报' + str(i + 1))
    else:
        label.extend(member_list)

    return_dict = {}
    return_dict["far"] = {}
    return_dict["pod"] = {}

    for line in range(new_hfmc_array_shape[0]):
        pofd = [1]
        pofd.extend(pofd_hfmc(new_hfmc_array[line, :]).tolist())
        pofd.append(0)
        pod = [1]
        pod.extend(pod_hfmc(new_hfmc_array[line, :]).tolist())
        pod.append(0)
        pofd = np.array(pofd)
        pod = np.array(pod)
        if (pofd.size < 30):
            plt.plot(pofd, pod,  linewidth=2, marker=".", label=label[line])
        else:
            plt.plot(pofd, pod,  linewidth=2, label=label[line])
        return_dict["far"][line] = pofd
        return_dict["pod"][line] = pod

    plt.plot([0, 1], [0, 1], ":", color="k", linewidth=1, label="无技巧")
    plt.xlabel("报空率(pofd)", fontsize=sup_fontsize * 0.9)
    plt.ylabel("命中率(pod)", fontsize=sup_fontsize * 0.9)
    plt.ylim(0.0, 1.0)
    plt.xlim(0.0, 1.0)
    plt.legend(loc=4, fontsize=sup_fontsize * 0.9)
    plt.title(title, fontsize=sup_fontsize)
    plt.xticks(fontsize=sup_fontsize * 0.8)
    plt.yticks(fontsize=sup_fontsize * 0.8)

    if save_path is None:
        show = True
    else:
        plt.savefig(save_path,bbox_inches='tight')
        print("检验结果已以图片形式保存至" + save_path)
    if show is True:
        plt.show()
    plt.close()
    return  return_dict

def roc_hnh(hnh_array,  member_list=None, save_path=None,show  =False, dpi = 300, title="ROC图",
            sup_fontsize =10,width = None,height = None):
    '''

    :param th_array:
    :param save_path:
    :return:
    '''
    hfmc_list = []
    grade_count = hnh_array.shape[-2]
    grade = 1 / grade_count
    if grade_count < 1:
        print('grade_count输入错误，不能小于1')
        return
    grade_list = np.arange(0, 1, grade).tolist()
    grade_list.append(1.1)
    shape = list(hnh_array.shape)
    hnh_array = hnh_array.reshape((-1, len(grade_list) - 1, 2))
    new_hnh_array_shape = hnh_array.shape
    for line in range(new_hnh_array_shape[0]):
        total_grade_num = hnh_array[line, :, 0]
        observed_grade_num = hnh_array[line, :, 1]
        ngrade = len(total_grade_num)
        hfmc = np.zeros((len(total_grade_num), 4))
        total_hap = np.sum(observed_grade_num)
        total_num = np.sum(total_grade_num)
        for i in range(ngrade):
            hfmc[i, 0] = np.sum(observed_grade_num[i:])
            hfmc[i, 1] = np.sum(total_grade_num[i:]) - hfmc[i, 0]
            hfmc[i, 2] = total_hap - hfmc[i, 0]
            hfmc[i, 3] = total_num - (hfmc[i, 0] + hfmc[i, 1] + hfmc[i, 2])
        hfmc_list.append(hfmc)
    hfmc_array = np.array(hfmc_list)
    shape = shape[:-2]
    shape.append(len(grade_list) - 1)
    shape.append(4)
    hfmc_array.reshape(shape)
    return_dict = roc_hfmc(hfmc_array,  member_list=member_list,dpi=dpi,show=show, save_path=save_path, title=title,sup_fontsize=sup_fontsize,width= width,height= height)
    return return_dict

def discrimination(Ob, Fo, grade_count=10,member_list=None,  vmax = None,log_y  = False, save_path=None,show = False,dpi = 300, title="区分能力图",
                    sup_fontsize =10,width = None,height = None):
    '''

    :param Ob:
    :param Fo:
    :param grade_count:
    :param save_path:
    :return:
    '''
    hnh_array = meteva.method.hnh(Ob, Fo, grade_count)
    discrimination_hnh(hnh_array, member_list = member_list,vmax = vmax,log_y = log_y, save_path=save_path,show = show,dpi = dpi, title=title,
                       width=width, height=height, sup_fontsize=sup_fontsize)


def discrimination_hnh(hnh_array,  member_list=None, vmax = None,log_y = False,save_path=None,show = False,dpi = 300, title="区分能力图",
                        sup_fontsize =10,width = None,height = None):
    '''

    :param th_array:
    :param save_path:
    :return:
    '''

    grade_count = hnh_array.shape[-2]
    grade = 1 / grade_count
    if grade_count < 1:
        print('grade_count输入错误，不能小于1')
        return
    grade_list = (np.arange(0, 1+0.5* grade, grade) *100).astype(np.int16)/100
    new_th_array = hnh_array.reshape((-1, len(grade_list) - 1, 2))
    new_th_array_shape = new_th_array.shape

    if width is None:
        width = meteva.base.plot_tools.caculate_axis_width(grade_list,sup_fontsize,new_th_array_shape[0])
        if width > 10:
            width = 10
        if width<5:
            width =5
    if height is None:
        height = width *0.5
    label = []
    legend_num = new_th_array_shape[0]
    if member_list is None:
        if legend_num == 1:
            label.append('预报')
        else:
            for i in range(legend_num):
                label.append('预报' + str(i + 1))
    else:
        label.extend(member_list)

    fig = plt.figure(figsize=(width, height),dpi = dpi)
    ymax = 0
    color_list = meteva.base.tool.color_tools.get_color_list(legend_num)
    legend_col = int(width * 0.7)
    legend_row = int(math.ceil(legend_num / legend_col))
    legend_col = int(math.ceil(legend_num / legend_row))

    axes = plt.subplot(1, 1, 1)
    grade = 1 / grade_count
    x = np.arange(grade / 2, 1, grade)
    bar_width = 1 / (grade_count * (legend_num+3))

    mark_line_x = []
    mark_line_y = []
    for line in range(new_th_array_shape[0]):
        total_grade_num = new_th_array[line, :, 0]
        observed_grade_num = new_th_array[line, :, 1]
        total_num = np.sum(total_grade_num)
        observed_grade_rate = observed_grade_num / total_num
        not_observed_grade_num = total_grade_num - observed_grade_num
        not_observed_grade_rate = not_observed_grade_num / total_num

        x1 = x + (line - legend_num/2 + 0.5) * bar_width
        mark_line_x.append(x1)
        mark_line_y.append((not_observed_grade_rate + observed_grade_rate)* x)

        axes.bar(x1, observed_grade_rate, width=bar_width*0.8,fc = color_list[line])
        axes.bar(x1, not_observed_grade_rate + observed_grade_rate, width=bar_width*0.8, fill=False,ec = color_list[line])

        ymax = max(np.max(not_observed_grade_rate + observed_grade_rate), ymax)
    mark_line_x = np.array(mark_line_x)
    mark_line_y = np.array(mark_line_y)
    mark_line_x = mark_line_x.T.flatten()
    mark_line_y = mark_line_y.T.flatten()
    axes.plot(mark_line_x, mark_line_y, '.', color='k')
    lines = axes.get_children()
    plt.xlabel("预测的概率", fontsize=sup_fontsize * 0.9)
    plt.ylabel("占总样本数的比例", fontsize=sup_fontsize * 0.9)
    if log_y: plt.yscale("log")

    if vmax is None:
        if log_y:
            vmax = ymax * 10
        else:
            vmax = ymax * 1.5


    plt.ylim(0.0, vmax)
    plt.xlim(0.0, 1)
    nlabel = len(label)
    legend1 = plt.legend([lines[0], lines[grade_count],lines[grade_count * nlabel * 2 ]], ['观测正例', '观测负例',"合理比例"],
                         loc='upper center', bbox_to_anchor=(0.5, 1), ncol=3,fontsize = sup_fontsize * 0.9)
    axes.add_artist(legend1)
    if legend_num > 1:
        legend2 = plt.legend(lines[0:grade_count * nlabel * 2:grade_count * 2], label, loc='upper center',
                             bbox_to_anchor=(0.5, 1 - 0.3/height), ncol = legend_col)

        axes.add_artist(legend2)

    plt.yticks(fontsize = sup_fontsize * 0.8)
    xtick = np.arange(0, 1.001, 0.1)
    plt.xticks(xtick, fontsize=sup_fontsize * 0.8)
    plt.title(title, fontsize=sup_fontsize)
    if save_path is None:
        show = True
    else:
        plt.savefig(save_path,bbox_inches='tight')
        print("检验结果已以图片形式保存至" + save_path)
    if show is True:
        plt.show()
    plt.close()


def comprehensive_probability(Ob, Fo, grade_count=10, member_list=None,vmax = None,log_y = False,save_path=None,dpi = 300,show = False, title="概率预报综合检验图",
                                sup_fontsize =10,width = None,height = None):
    '''
    :param Ob:
    :param Fo:
    :param save_path:
    :return:
    '''
    hnh_array = meteva.method.hnh(Ob, Fo, grade_count)
    comprehensive_hnh(hnh_array,member_list = member_list,vmax = vmax,log_y = log_y,save_path=save_path,show = show, dpi = dpi,title=title,
                      width=width, height=height, sup_fontsize=sup_fontsize)


def comprehensive_hnh(hnh_array,  member_list=None,vmax = None,log_y = False, save_path=None,show = False,dpi = 300, title="概率预报综合检验图",
                      sup_fontsize =10,width = None,height = None):
    '''

    :param th_array:
    :param save_path:
    :return:
    '''
    if width is None:
        width  = 8
    if height is None:
        height = 5.6
    fig = plt.figure(figsize=(width,height),dpi = dpi)
    grade_count = hnh_array.shape[-2]
    grade = 1 / grade_count
    if grade_count < 1:
        print('grade_count输入错误，不能小于1')
        return
    grade_list = np.arange(0, 1, grade).tolist()
    grade_list.append(1.1)
    shape = list(hnh_array.shape)
    new_th_array = hnh_array.reshape((-1, len(grade_list) - 1, 2))
    new_th_array_shape = new_th_array.shape
    label = []
    legend_num = new_th_array_shape[0]
    if member_list is None:
        if legend_num == 1:
            label.append('预报')
        else:
            for i in range(legend_num):
                label.append('预报' + str(i + 1))
    else:
        label.extend(member_list)
    color_list = meteva.base.tool.color_tools.get_color_list(legend_num)
    bar_width = 1.0 / (grade_count * (legend_num + 3))
    ymax = -9999
    mark_line_x = []
    mark_line_y = []
    for line in range(legend_num):
        total_grade_num = new_th_array[line, :, 0]
        observed_grade_num = new_th_array[line, :, 1]
        total_num = np.sum(total_grade_num)

        under = np.zeros_like(total_grade_num)
        under[:] = total_grade_num[:]
        under[total_grade_num == 0] = 1
        ob_rate = observed_grade_num / under
        ob_rate[total_grade_num == 0] = IV
        ob_rate_noIV = set_plot_IV(ob_rate)
        ob_rate[total_grade_num == 0] = np.nan
        index_iv = np.where(total_grade_num == 0)

        not_observed_grade_num = total_grade_num - observed_grade_num

        ngrade = len(total_grade_num)
        grade = 1 / ngrade
        x = np.arange(grade / 2, 1, grade)

        line_x = np.arange(0, 1.01, 0.1)
        prefect_line_y = np.arange(0, 1.01, 0.1)
        climate_line_y = np.ones_like(line_x) * np.sum(observed_grade_num) / total_num

        grid_plt = plt.GridSpec(6, 2)
        if line == 0:
            plt.suptitle(title,fontsize= sup_fontsize,y=0.95)
            plt.subplots_adjust(wspace=0.2, hspace=1)
            ax3 = plt.subplot(grid_plt[0:4, 1])
            ax4 = plt.subplot(grid_plt[4:, :])
            ax1 = plt.subplot(grid_plt[0:4, 0])
        ax1.plot(x, ob_rate_noIV, "--", linewidth=0.5, color="k")
        x_iv = x[index_iv[0]]
        ob_rate_noIV_iv = ob_rate_noIV[index_iv[0]]

        ax1.plot(x_iv, ob_rate_noIV_iv, "x", color='k',label = None)
        if line == 0:
            ax1.plot(line_x, prefect_line_y, '--', label="完美", color="k")
            ax1.plot(line_x, climate_line_y, ':', label="无技巧", color="k")
        ax1.plot(x, ob_rate, marker=".", markersize="10", color=color_list[line],label = None)

        x1 = x + (line -legend_num / 2 + 0.5) * bar_width
        total_hap = np.sum(observed_grade_num)
        hfmc = np.zeros((len(total_grade_num), 4))
        for i in range(ngrade):
            hfmc[i, 0] = np.sum(observed_grade_num[i:])
            hfmc[i, 1] = np.sum(total_grade_num[i:]) - hfmc[i, 0]
            hfmc[i, 2] = total_hap - hfmc[i, 0]
            hfmc[i, 3] = total_num - (hfmc[i, 0] + hfmc[i, 1] + hfmc[i, 2])

        pofd = [1]
        pofd.extend(pofd_hfmc(hfmc).tolist())
        pofd.append(0)
        pod = [1]
        pod.extend(pod_hfmc(hfmc).tolist())
        pod.append(0)
        far = np.array(pofd)
        pod = np.array(pod)

        if line == 0:
            ax3.plot([0, 1], [0, 1], ":", color="k", linewidth=1, label="无技巧")
        ax3.plot(far, pod, color=color_list[line], linewidth=2, label=label[line])
        #ax4.bar(x1, observed_grade_num, width=bar_width*0.8, color=color_list[line],label = label[line])
        ax4.bar(x1, (not_observed_grade_num+ observed_grade_num), width=bar_width*0.8,edgecolor=color_list[line],label =label[line])

        ax4.set_xlabel("预报概率",fontsize= sup_fontsize *0.9)
        ax4.set_ylabel("样本数",fontsize= sup_fontsize *0.9)
        ymax = max(np.max(observed_grade_num+not_observed_grade_num), ymax)
        mark_line_x.append(x1)
        mark_line_y.append((not_observed_grade_num+ observed_grade_num)* x)

    ax1.set_xlim(0.0, 1)
    ax1.set_ylim(0.0, 1)
    ax1.set_xlabel("预报概率",fontsize= sup_fontsize *0.9)
    ax1.set_ylabel("观测频率",fontsize= sup_fontsize *0.9)
    ax1.legend(loc=2,fontsize= sup_fontsize *0.9)
    ax3.set_xlabel("报空率(pofd)",fontsize= sup_fontsize *0.9)
    ax3.set_ylabel("命中率(pod)",fontsize= sup_fontsize *0.9)
    ax3.set_ylim(0.0, 1.0)
    ax3.set_xlim(0.0, 1.0)
    ax3.legend(loc=4,fontsize= sup_fontsize *0.9)
    ax4.set_xlim(0.0, 1)
    ax4.set(title='\n')
    ax4.set_xticks(np.arange(0,1.01,1/grade_count))
    #ax4.legend(loc="upper right", ncol=2)

    #mark_line_x = np.array(mark_line_x)
    #mark_line_y = np.array(mark_line_y)
    #mark_line_x = mark_line_x.T.flatten()
    #mark_line_y = mark_line_y.T.flatten()
    #ax4.plot(mark_line_x, mark_line_y, '.', color='k',markersize = bar_width * 300)
    #lines = ax4.get_children()
    #ax4.legend([lines[0], lines[grade_count],lines[grade_count * legend_num * 2 ]], ['观测正例', '观测负例',"合理比例"],
    #                     loc='upper center', bbox_to_anchor=(0.5, 1.0), ncol=3, prop={'size': 6},fontsize= sup_fontsize *0.9)

    if log_y: ax4.set_yscale('log')
    if vmax is None:
        ax4.set_ylim(0.0, ymax * 1.5)
    else:
        ax4.set_ylim(0.0, vmax)
    if save_path is None:
        show = True
    else:
        plt.savefig(save_path,bbox_inches='tight')
        print("检验结果已以图片形式保存至" + save_path)
    if show is True:
        plt.show()
    plt.close()
